from __future__ import print_function, unicode_literals

from os.path import dirname, join as ospj
from random import randint
import unittest
import sys
if sys.version_info < (3, 0): 
    from SimpleXMLRPCServer import SimpleXMLRPCServer
    from xmlrpclib import ServerProxy
else:
    from xmlrpc.server import SimpleXMLRPCServer
    from xmlrpc.client import ServerProxy
from tempfile import mkdtemp
from threading import Lock, Thread, local as thread_local

from DenyHosts.constants import SYNC_HOSTS, SYNC_RECEIVED_HOSTS, SYNC_TIMESTAMP
from DenyHosts.prefs import Prefs
from DenyHosts.sync import Sync

LOCAL_SYNC_SERVER_ADDRESS = ('127.0.0.1', 9911)
LOCAL_SYNC_SERVER_URL = 'http://%s:%d' % LOCAL_SYNC_SERVER_ADDRESS

class MockSyncServer(object):
    def __init__(self):
        self.hosts = []

    def get_hosts(self):
        """
        Test method to directly return stored hosts
        """
        return self.hosts

    def get_new_hosts(self, timestamp, threshold, hosts_added, download_resiliency):
        """
        Matches the API inferred from the Sync class
        """
        return {
            'hosts': self.hosts,
            'timestamp': '0',
        }

    def add_hosts(self, hosts):
        self.hosts.extend(hosts)

class SyncServerTest(unittest.TestCase):
    """
    Base class of all Sync test classes that use a mock sync server.
    """
    def sync_server(self):
        self.thread_local.alive = True
        server = None
        try:
            server = SimpleXMLRPCServer(LOCAL_SYNC_SERVER_ADDRESS, allow_none=True, logRequests=False)
            server.register_function(self._exit, 'exit')
            sync_server = MockSyncServer()
            server.register_instance(sync_server)
        finally:
            self.lock.release()
        if server is not None:
            while True:
                if not self.thread_local.alive:
                    break
                server.handle_request()

    def _exit(self):
        self.thread_local.alive = False

    def setUp(self):
        # Poor-man's version of a threading.Barrier (which was added in 3.2). Make a
        # new threading.Lock and immediately acquire it. Spawn the sync_server thread,
        # then acquire the lock again. The second acquire() call will block until the
        # sync server thread is initialized -- it calls release() before entering the
        # serve_forever loop. This isn't totally free of race conditions -- a test
        # method could conceivably be run between sync_server's calls to
        # self.lock.release() and server.serve_forever. We just need our test runs
        # to behave as deterministically as possible, so hopefully this is good
        # enough for now.
        # TODO: use a real threading.Barrier when we drop support for Python 2
        self.lock = Lock()
        self.lock.acquire()
        self.thread_local = thread_local()
        self.server_thread = Thread(target=self.sync_server)
        self.server_thread.start()
        self.lock.acquire()
        self.remote_sync_server = ServerProxy(LOCAL_SYNC_SERVER_URL, allow_none=True)
        self.lock.release()

    def tearDown(self):
        self.remote_sync_server.exit()
        self.server_thread.join()

class MockSyncServerTest(SyncServerTest):
    """
    Test the mock sync server itself.
    """
    def test_add_hosts(self):
        hosts = ['host1', 'host2']
        self.remote_sync_server.add_hosts(['host1', 'host2'])
        data = self.remote_sync_server.get_new_hosts(None, None, None, None)
        self.assertEqual(data['hosts'], hosts)
        self.assertEqual(self.remote_sync_server.get_hosts(), hosts)

    def test_add_no_hosts(self):
        self.remote_sync_server.add_hosts([])
        data = self.remote_sync_server.get_new_hosts(None, None, None, None)
        self.assertFalse(data['hosts'])
        self.assertFalse(self.remote_sync_server.get_hosts())

class SyncTestStaticTimestamp(unittest.TestCase):
    """
    Tests that we can read the sync timestamp from the filesystem.
    """
    def setUp(self):
        self.prefs = Prefs()
        self.prefs._Prefs__data['WORK_DIR'] = ospj(dirname(__file__), 'data/sync/static')
        self.sync = Sync(self.prefs)

    def test_get_sync_timestamp(self):
        timestamp = 427850432
        self.assertEqual(self.sync.get_sync_timestamp(), timestamp)

class SyncTestDynamicTimestamp(unittest.TestCase):
    """
    Tests that we can set the timestamp on the filesystem. Separated
    into a different test class to avoid clobbering the static test data
    for SyncTestStaticTimestamp.
    """
    def setUp(self):
        self.prefs = Prefs()
        self.prefs._Prefs__data['WORK_DIR'] = ospj(dirname(__file__), 'data/sync/dynamic')
        self.sync = Sync(self.prefs)
        self.value = randint(0, 1e9)

    def test_set_sync_timestamp(self):
        self.sync.set_sync_timestamp(str(self.value))
        path = ospj(self.prefs._Prefs__data['WORK_DIR'], SYNC_TIMESTAMP)
        with open(path) as f:
            saved_timestamp = int(f.read().strip())
        self.assertEqual(self.value, saved_timestamp)

class SyncTestBasic(SyncServerTest):
    def setUp(self):
        super(SyncTestBasic, self).setUp()
        self.prefs = Prefs()
        self.prefs._Prefs__data['SYNC_SERVER'] = LOCAL_SYNC_SERVER_URL
        self.prefs._Prefs__data['WORK_DIR'] = ospj(dirname(__file__), 'data/sync/static')

    def test_connect_disconnect(self):
        sync = Sync(self.prefs)
        self.assertTrue(sync.xmlrpc_connect())
        self.assertFalse(sync._Sync__server is None)
        self.assertTrue(sync._Sync__connected)
        sync.xmlrpc_disconnect()
        self.assertTrue(sync._Sync__server is None)
        self.assertFalse(sync._Sync__connected)

class SyncTestSendHosts(SyncServerTest):
    def setUp(self):
        super(SyncTestSendHosts, self).setUp()
        self.prefs = Prefs()
        self.prefs._Prefs__data['SYNC_SERVER'] = LOCAL_SYNC_SERVER_URL
        work_dir = mkdtemp()
        self.prefs._Prefs__data['WORK_DIR'] = work_dir

        self.test_hosts = ['host1', 'host2', 'host3']
        with open(ospj(work_dir, SYNC_HOSTS), 'w') as f:
            for host in self.test_hosts:
                print(host, file=f)

    def test_send_hosts(self):
        sync = Sync(self.prefs)
        self.assertFalse(sync._Sync__hosts_added)
        self.assertTrue(sync.send_new_hosts())
        self.assertEqual(sync._Sync__hosts_added, self.test_hosts)

class SyncTestReceiveHosts(SyncServerTest):
    def setUp(self):
        super(SyncTestReceiveHosts, self).setUp()
        self.prefs = Prefs()
        self.prefs._Prefs__data['SYNC_SERVER'] = LOCAL_SYNC_SERVER_URL
        self.work_dir = mkdtemp()
        self.prefs._Prefs__data['WORK_DIR'] = self.work_dir
        self.test_hosts = ['host1', 'host2', 'host3']
        self.remote_sync_server.add_hosts(self.test_hosts)

    def test_send_hosts(self):
        sync = Sync(self.prefs)
        self.assertEqual(sync.receive_new_hosts(), self.test_hosts)
        filename = ospj(self.work_dir, SYNC_RECEIVED_HOSTS)
        with open(filename) as f:
            hosts = [line.strip().split(':')[0] for line in f]
        self.assertEqual(self.test_hosts, hosts)



#!/usr/bin/env python
import os, sys

def usage():
    print "%s WORK_DIR [num_results]" % sys.argv[0]
    sys.exit(1)

try:
    work_dir = sys.argv[1]
except IndexError:
    print "you must specify your DenyHosts WORK_DIR"
    usage()

try:
    num = int(sys.argv[2])
except IndexError:
    num = 10

fname = os.path.join(work_dir, "users-invalid")

try:
    fp = open(fname, "r")
except IOError:
    print fname, "does not exist"
    sys.exit(1)

d = {}

for line in fp:
    try:
        foo = line.split(":")
        username = foo[0]
        attempts = int(foo[1])
        # timestamp = foo[2].strip()
    except Exception:
        continue

    l = d.get(attempts, [])
    l.append(username)
    d[attempts] = l

fp.close()

keys = d.keys()
keys.sort()
keys.reverse()

i = 0
for key in keys:
    l = d.get(key)
    for username in l:
        i += 1
        print username
        if i >= num: break
    if i >= num: break



import sys

#################################################################################
#        These files will be created relative to prefs WORK_DIR                 #
#################################################################################

SECURE_LOG_OFFSET = "offset"
DENIED_TIMESTAMPS = "denied-timestamps"

ABUSIVE_HOSTS_INVALID = "hosts"
ABUSIVE_HOSTS_VALID = "hosts-valid"
ABUSIVE_HOSTS_ROOT = "hosts-root"
ABUSIVE_HOSTS_RESTRICTED = "hosts-restricted"

ABUSED_USERS_INVALID = "users-invalid"
ABUSED_USERS_VALID = "users-valid"
ABUSED_USERS_AND_HOSTS = "users-hosts"
SUSPICIOUS_LOGINS = "suspicious-logins"   # successful logins AFTER invalid
                                          #   attempts from same host

ALLOWED_HOSTS = "allowed-hosts"
ALLOWED_WARNED_HOSTS = "allowed-warned-hosts"

RESTRICTED_USERNAMES = "restricted-usernames"

SYNC_TIMESTAMP = "sync-timestamp"
SYNC_HOSTS = "sync-hosts"
SYNC_HOSTS_TMP = "sync-hosts.tmp"
SYNC_RECEIVED_HOSTS = "sync-received"

PURGE_HISTORY = "purge-history"

#################################################################################
#                           Miscellaneous constants                             #
#################################################################################

CONFIG_FILE = "/etc/denyhosts.conf"

DENY_DELIMITER = "# DenyHosts:"
ENTRY_DELIMITER = " | "

TIME_SPEC_LOOKUP =  {
    's': 1,        # s
    'm': 60,       # minute
    'h': 3600,     # hour
    'd': 86400,    # day
    'w': 604800,   # week
    'y': 31536000, # year
}

SYNC_MIN_INTERVAL = 300 # 5 minutes

plat = sys.platform
if plat.startswith("freebsd"):
    # this has no effect if BLOCK_SERVICE is empty
    BSD_STYLE = " : deny"
else:
    BSD_STYLE = ""



import sys
from textwrap import dedent

MINIMUM_VERSION = (2, 4)

def check_version():
    if sys.version_info < MINIMUM_VERSION:
        min_version_str = '.'.join(str(x) for x in MINIMUM_VERSION)
        print("Python >= %s required.  You are using:\n%s" % (min_version_str, sys.version))

        print(dedent("""
            ######################################################################

            Visit http://www.python.org and download a more recent version of
            Python.

            You should install this version in addition to your current version
            (rather than upgrading your current version) because your system might
            depend on the current version.  After installing the newer version, for
            instance version 3.2, simply invoke DenyHosts explicitly with the new
            version of python, eg:

            $ python3.4 %s

            ######################################################################

            """) % ' '.join(sys.argv))
        sys.exit(1)

check_version()



import logging
import time

debug = logging.getLogger("counter").debug

# TODO define __eq__ for this class
class CounterRecord(object):
    def __init__(self, count=0, date=None):
        self.__count = count
        if not date:
            self.__date = time.asctime()
        else:
            self.__date = date

    def __str__(self):
        return "%d:%s" % (self.__count, self.__date)

    def __repr__(self):
        return "CountRecord <%d - %s>" % (self.__count, self.__date)

    def __add__(self, increment):
        # AAAAH!
        # Expressions like `a + 4` are usually assumed to not have any side effects,
        # but this is not the case with CounterRecord objects. With `c = CounterRecord()`,
        # simply evaluating `c + 1` will increment c.__count by 1. This is horrifying.
        self.__count += increment
        self.__date = time.asctime()
        return self

    def get_count(self):
        return self.__count

    def get_date(self):
        return self.__date

    def reset_count(self):
        self.__count = 0

    def age_count(self, age):
        cutoff = int(time.time()) - age
        epoch = time.mktime(time.strptime(self.__date))
        #debug("cutoff : %d", cutoff)
        #debug("epoch  : %d", epoch)
        if cutoff > epoch:
            self.__count = 0


# TODO replace this with collections.defaultdict
class Counter(dict):
    """
     Behaves like a dictionary, except that if the key isn't found, 0 is returned
     rather than an exception.  This is suitable for situations like:
         c = Counter()
         c['x'] += 1
    """
    def __init__(self):
        dict.__init__(self)

    def __getitem__(self, k):
        try:
            return dict.__getitem__(self, k)
        except KeyError:
            count_rec = CounterRecord(0)
            #debug("%s - %s", k, count_rec)
            self.__setitem__(k, count_rec)
            #debug("dict: %s", dict.values(self))
            #debug("count_rec: %s", count_rec)
            return count_rec



from __future__ import print_function, unicode_literals

import unittest

from DenyHosts.report import Report

class ReportTest(unittest.TestCase):
    def test_empty(self):
        report = Report(hostname_lookup='false')
        self.assertTrue(report.empty())
        self.assertEqual(report.get_report(), '')

    def test_add_section(self):
        report = Report(hostname_lookup='false')
        report.add_section('message', ['line1', ('line2', 1)])
        expected_report = ('message:\n\nline1\nline2: 1\n\n\n------------------'
            '----------------------------------------------------\n')
        self.assertEqual(report.get_report(), expected_report)



import logging
import logging.handlers
from smtplib import SMTP
from smtplib import SMTPResponseException
from smtplib import SMTPHeloError
import sys
from textwrap import dedent
import time
import ipaddr

from .constants import BSD_STYLE, TIME_SPEC_LOOKUP
from .regex import TIME_SPEC_REGEX

debug = logging.getLogger("util").debug

def setup_logging(prefs, enable_debug, verbose, daemon):
        daemon_log = prefs.get('DAEMON_LOG')
        if daemon_log:
            # define a Handler which writes INFO messages or higher to the sys.stderr
            # fh = logging.FileHandler(daemon_log, 'a')
            fh = logging.handlers.RotatingFileHandler(daemon_log, 'a', 1024*1024, 7)
            fh.setLevel(logging.DEBUG)
            formatter = logging.Formatter(prefs.get('DAEMON_LOG_MESSAGE_FORMAT'),
                prefs.get('DAEMON_LOG_TIME_FORMAT'))
            fh.setFormatter(formatter)
            # add the handler to the root logger
            logging.getLogger().addHandler(fh)
            if enable_debug:
                # if --debug was enabled provide gory activity details
                logging.getLogger().setLevel(logging.DEBUG)
                #prefs.dump_to_logger()
            else:
                # in daemon mode we always log some activity
                logging.getLogger().setLevel(logging.INFO)

            info = logging.getLogger("denyhosts").info
            info("DenyHosts launched with the following args:")
            info("   %s", ' '.join(sys.argv))
            prefs.dump_to_logger()

def die(msg, ex=None):
    print(msg)
    if ex:
        print(ex)
    sys.exit(1)

def is_true(s):
    return s.lower() in ('1', 't', 'true', 'y', 'yes')

def is_false(s):
    return not is_true(s)

def calculate_seconds(timestr, zero_ok=False):
    # return the number of seconds in a given timestr such as 1d (1 day),
    # 13w (13 weeks), 5s (5seconds), etc...
    if type(timestr) is int: return timestr

    m = TIME_SPEC_REGEX.search(timestr)
    if not m:
        raise Exception("Invalid time specification: string format error: %s", timestr)

    units = int(m.group('units'))
    period = m.group('period') or 's' # seconds is the default

    if units == 0 and not zero_ok:
        raise Exception("Invalid time specification: units = 0")

    seconds = units * TIME_SPEC_LOOKUP[period]
    #info("converted %s to %ld seconds: ", timestr, seconds)
    return seconds

def parse_host(line):
    # parses a line from /etc/hosts.deny
    # returns the ip address

    # the deny file can be in the form:
    # 1) ip_address
    # 2) sshd: ip_address
    # 3) ip_address : deny
    # 4) sshd: ip_address : deny

    # convert form 3 & 4 to 1 & 2
    try:
        vals = line.split(":")

        # we're only concerned about the ip_address
        if len(vals) == 1:
            form = vals[0]
        else:
            form = vals[1]

        host = form.strip()
    except Exception:
        host = ""
    return host

def send_email(prefs, report_str):
    recipients = prefs['ADMIN_EMAIL'].split(',')

    msg = dedent("""
        From: %s
        To: %s
        Subject: %s
        Date: %s

        """).lstrip() % (
            prefs.get('SMTP_FROM'),
            prefs.get('ADMIN_EMAIL'),
            prefs.get('SMTP_SUBJECT'),
            time.strftime(prefs.get('SMTP_DATE_FORMAT')
        )
    )

    msg += report_str
    try:
        smtp = SMTP()

        if logging.getLogger().isEnabledFor(logging.DEBUG):
            smtp.set_debuglevel(1)

        smtp.connect(prefs.get('SMTP_HOST'),
            prefs.get('SMTP_PORT'))

        # If the server supports ESMTP and TLS, then convert the message exchange to TLS via the
        # STARTTLS command.
        if smtp.ehlo()[0] == 250:
            if smtp.has_extn('starttls'):
                (code, resp) = smtp.starttls()
                if code != 220:
                    raise SMTPResponseException(code, resp)
                (code, resp) = smtp.ehlo()
                if code != 250:
                    raise SMTPResponseException(code, resp)
        else:
            # The server does not support esmtp.

            # The Python library SMTP class handles executing HELO/EHLO commands inside
            # login/sendmail methods when neither helo()/ehlo() methods have been
            # previously called.  Because we have already called ehlo() above, we must
            # manually fallback to calling helo() here.

            (code, resp) = smtp.helo()
            if not (200 <= code <= 299):
                raise SMTPHeloError(code, resp)

        username = prefs.get('SMTP_USERNAME')
        password = prefs.get('SMTP_PASSWORD')

        if username and password:
            smtp.login(username, password)

        smtp.sendmail(prefs.get('SMTP_FROM'),
                      recipients,
                      msg)
        debug("sent email to: %s" % prefs.get("ADMIN_EMAIL"))
    except Exception as e:
        print("Error sending email")
        print(e)
        print("Email message follows:")
        print(msg)

    try:
        smtp.quit()
    except Exception:
        pass

def normalize_whitespace(string):
    return ' '.join(string.split())

def is_valid_ip_address(ip_address):
    try:
        ip = ipaddr.IPAddress(ip_address)
    except:
        return False
    if (ip.is_reserved or ip.is_private or ip.is_loopback or
        ip.is_unspecified or ip.is_multicast or
        ip.is_link_local):
        return False
    return True



import os
import logging
from .constants import SECURE_LOG_OFFSET

debug = logging.getLogger("filetracker").debug

class FileTracker(object):
    def __init__(self, work_dir, logfile):
        self.work_dir = work_dir
        self.logfile = logfile
        (self.__first_line, self.__offset) = self.__get_current_offset()

    def __get_last_offset(self):
        path = os.path.join(self.work_dir,
                            SECURE_LOG_OFFSET)
        first_line = ""
        offset = 0
        try:
            fp = open(path, "r")
            first_line = fp.readline()[:-1]
            offset = int(fp.readline())
        except IOError:
            pass

        debug("__get_last_offset():")
        debug("   first_line: %s", first_line)
        debug("   offset: %ld", offset)

        return first_line, offset

    def __get_current_offset(self):
        first_line = ""
        offset = 0
        try:
            fp = open(self.logfile, "r")
            first_line = fp.readline()[:-1]
            fp.seek(0, 2)
            offset = fp.tell()
        except IOError as e:
            raise e

        debug("__get_current_offset():")
        debug("   first_line: %s", first_line)
        debug("   offset: %ld", offset)

        return first_line, offset

    def update_first_line(self):
        first_line = ""
        try:
            fp = open(self.logfile, "r")
            first_line = fp.readline()[:-1]
        except IOError as e:
            raise e

        self.__first_line = first_line

    def get_offset(self):
        last_line, last_offset = self.__get_last_offset()

        if last_line != self.__first_line:
            # log file was rotated, start from beginning
            offset = 0
        elif self.__offset > last_offset:
            # new lines exist in log file
            offset = last_offset
        else:
            # no new entries in log file
            offset = None

        debug("get_offset():")
        debug("   offset: %s", str(offset))

        return offset

    def save_offset(self, offset):
        path = os.path.join(self.work_dir,
                            SECURE_LOG_OFFSET)
        try:
            fp = open(path, "w")
            fp.write("%s\n" % self.__first_line)
            fp.write("%ld\n" % offset)
            fp.close()
        except IOError:
            print("Could not save logfile offset to: %s" % path)



from __future__ import print_function, unicode_literals

from os.path import abspath, dirname, join as ospj
from tempfile import mkdtemp
import unittest

from DenyHosts.allowedhosts import AllowedHosts
from DenyHosts.constants import ALLOWED_WARNED_HOSTS
from DenyHosts.prefs import Prefs

# TODO use assertIn/assertNotIn when dropping support for 2.6

class AllowedHostsBase(unittest.TestCase):
    def setUp(self):
        data_dir = ospj(dirname(abspath(__file__)), 'data')
        self.prefs = Prefs()
        self.prefs._Prefs__data['WORK_DIR'] = data_dir
        self.prefs._Prefs__data['ALLOWED_HOSTS_HOSTNAME_LOOKUP'] = 'false'
        self.allowed_hosts = AllowedHosts(self.prefs)

class AllowedHostsBasicTest(AllowedHostsBase):
    def test_positives(self):
        self.assertTrue('127.0.0.1' in self.allowed_hosts)
        self.assertTrue('172.16.0.1' in self.allowed_hosts)

    def test_negatives(self):
        self.assertFalse('10.0.0.1' in self.allowed_hosts)
        self.assertFalse('127.0.0.2' in self.allowed_hosts)
        self.assertFalse('0.0.0.0' in self.allowed_hosts)
        self.assertFalse('bogus' in self.allowed_hosts)

class AllowedHostsWildcardTest(AllowedHostsBase):
    def test_positives(self):
        self.assertTrue('192.168.1.1' in self.allowed_hosts)
        self.assertTrue('192.168.1.255' in self.allowed_hosts)

    def test_negatives(self):
        self.assertFalse('192.168.0.1' in self.allowed_hosts)
        self.assertFalse('255.255.255.255' in self.allowed_hosts)

class AllowedHostsRangeTest(AllowedHostsBase):
    def test_positives(self):
        self.assertTrue('1.1.1.20' in self.allowed_hosts)
        self.assertTrue('1.1.1.39' in self.allowed_hosts)

    def test_negatives(self):
        self.assertFalse('1.1.1.10' in self.allowed_hosts)
        self.assertFalse('1.1.1.50' in self.allowed_hosts)

class AllowedHostsHostnameTest(AllowedHostsBase):
    def test_positives(self):
        self.assertTrue('hostname' in self.allowed_hosts)

    def test_negatives(self):
        self.assertFalse('another_hostname' in self.allowed_hosts)

class AllowedHostsWarnedHostsTest(unittest.TestCase):
    """
    Not a subclass of AllowedHostsBase since testing the warned
    hosts functionality is kind of distinct from tracking a set
    of allowed hosts
    """
    def setUp(self):
        self.work_dir = mkdtemp()
        self.warned_hosts_filename = ospj(self.work_dir, ALLOWED_WARNED_HOSTS)
        self.prefs = Prefs()
        self.prefs._Prefs__data['WORK_DIR'] = self.work_dir
        self.prefs._Prefs__data['ALLOWED_HOSTS_HOSTNAME_LOOKUP'] = 'false'
        self.allowed_hosts = AllowedHosts(self.prefs)

    def test_no_new_warned_host(self):
        self.assertFalse(self.allowed_hosts.get_new_warned_hosts())

    def test_no_warned_host(self):
        self.allowed_hosts.load_warned_hosts()
        self.assertFalse(self.allowed_hosts.get_new_warned_hosts())

    def test_load_warned_hosts(self):
        hosts = ['host1', 'host2']
        with open(self.warned_hosts_filename, 'w') as f:
            for host in hosts:
                print(host, file=f)

        self.allowed_hosts.load_warned_hosts()
        self.assertEqual(set(hosts), set(self.allowed_hosts.warned_hosts))

    def test_save_warned_hosts(self):
        hosts = ['host1', 'host2']
        test_string = ''.join(host + '\n' for host in hosts)

        for host in hosts:
            self.allowed_hosts.add_warned_host(host)
        self.allowed_hosts.save_warned_hosts()

        with open(self.warned_hosts_filename) as f:
            self.assertEqual(test_string, f.read())



import logging
import re
import socket
# Removing this as it causes runtime errors on Python3.4
# from types import ListType, TupleType
import types
from .util import is_true
try:
    import syslog
    HAS_SYSLOG = True
except ImportError:
    HAS_SYSLOG = False

debug = logging.getLogger("report").debug
warn = logging.getLogger("report").warn

IP_ADDR_REGEX = re.compile(r"""(?P<ip>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})""")

class Report:
    def __init__(self, hostname_lookup, use_syslog=False):
        self.report = ""
        if use_syslog and not HAS_SYSLOG:
            warn("syslog is unavailable on this platform")
        self.use_syslog = use_syslog and HAS_SYSLOG
        if self.use_syslog:
            syslog.openlog("denyhosts")
        self.hostname_lookup = is_true(hostname_lookup)

    def empty(self):
        if self.report: return 0
        else: return 1

    def clear(self):
        self.report = ""

    def get_report(self):
        return self.report

    def add_section(self, message, iterable):
        self.report += "%s:\n\n" % message
        for i in iterable:
            # if type(i) in (TupleType, ListType):
            if (type(i) is types.ListType) or (type(i) is types.TupleType):
                extra = ": %d\n" % i[1]
                i = i[0]
            else:
                extra = ""
            if self.hostname_lookup:
                hostname = self.get_hostname(i)
                debug("get_host: %s", hostname)
            else: hostname = i

            self.report += "%s%s\n" % (hostname, extra)

            if self.use_syslog:
                syslog.syslog("%s - %s%s" %(message, hostname, extra))
        self.report += "\n" + "-" * 70 + "\n"


    def get_hostname(self, text):
        m = IP_ADDR_REGEX.search(text)

        if m:
            start = m.start()
            ip = m.group('ip')
            text = text[:start]
        else:
            return text

        hostname = socket.getfqdn(ip)
        if hostname == ip:
            hostname = "unknown"
        return "%s (%s)" % (ip, hostname)



from __future__ import print_function, unicode_literals

from os.path import dirname, join as ospj
import unittest

from DenyHosts.deny_hosts import DenyHosts
from DenyHosts.lockfile import LockFile
from DenyHosts.prefs import Prefs

class DenyHostsBasicTest(unittest.TestCase):
    def setUp(self):
        self.directory = ospj(dirname(__file__), 'data/deny_hosts')
        self.work_dir = ospj(self.directory, 'work')
        self.logfile = ospj(self.work_dir, 'logfile')
        self.prefs = Prefs()

        self.lock_file = LockFile(ospj(self.directory, 'lockfile'))
        self.lock_file.remove(die_=False)
        self.lock_file.create()

        self.prefs._Prefs__data['ETC_DIR'] = ospj(self.directory, 'etc')
        self.prefs._Prefs__data['WORK_DIR'] = self.work_dir

    def test_init(self):
        DenyHosts(self.logfile, self.prefs, self.lock_file)

    def tearDown(self):
        self.lock_file.remove()



import re

#################################################################################
# REGULAR EXPRESSIONS ARE COOL.  Check out Kodos (http://kodos.sourceforge.net) #
#################################################################################

#DATE_FORMAT_REGEX = re.compile(r"""(?P<month>[A-z]{3,3})\s*(?P<day>\d+)""")

SSHD_FORMAT_REGEX = re.compile(r""".* (sshd.*:|\[sshd\]) (?P<message>.*)""")
#SSHD_FORMAT_REGEX = re.compile(r""".* sshd.*: (?P<message>.*)""")

FAILED_ENTRY_REGEX = re.compile(r"""Failed (?P<method>\S*) for (?P<invalid>invalid user |illegal user )?(?P<user>.*) from (::ffff:)?(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})( port \d+)?( ssh2)?$""")

FAILED_ENTRY_REGEX2 = re.compile(r"""(?P<invalid>(Illegal|Invalid)) user (?P<user>.*) from (::ffff:)?(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$""")

FAILED_ENTRY_REGEX3 = re.compile(r"""Authentication failure for (?P<user>.*) .*from (::ffff:)?(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})""")

FAILED_ENTRY_REGEX4 = re.compile(r"""Authentication failure for (?P<user>.*) .*from (?P<host>.*)""")

FAILED_ENTRY_REGEX5 = re.compile(r"""User (?P<user>.*) .*from (?P<host>.*) not allowed because none of user's groups are listed in AllowGroups$""")

FAILED_ENTRY_REGEX6 = re.compile(r"""Did not receive identification string .*from (::ffff:)?(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})""")

FAILED_ENTRY_REGEX7 = re.compile(r"""User (?P<user>.*) .*from (::ffff:)?(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) not allowed because not listed in AllowUsers""")

FAILED_ENTRY_REGEX8 = re.compile(r"""authentication error for (?P<user>.*) .*from (?P<host>.*)""")
FAILED_ENTRY_REGEX9 = re.compile(r"""Invalid user (?P<user>.*) .*from (::ffff:)?(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})""")

# these are reserved for future versions
FAILED_ENTRY_REGEX10 = None

# this should match the highest num failed_entry_regex + 1
FAILED_ENTRY_REGEX_NUM = 10 

FAILED_ENTRY_REGEX_RANGE = list(range(1, FAILED_ENTRY_REGEX_NUM))
FAILED_ENTRY_REGEX_MAP = {}

# create a hash of the failed entry regex'es indexed from 1 .. FAILED_ENTRY_REGEX_NUM
for i in FAILED_ENTRY_REGEX_RANGE:
    if i == 1: extra = ""
    else: extra = "%i" % i
    rx = eval("FAILED_ENTRY_REGEX%s" % extra)
    FAILED_ENTRY_REGEX_MAP[i] = rx


SUCCESSFUL_ENTRY_REGEX = re.compile(r"""Accepted (?P<method>\S+) for (?P<user>.*) from (::ffff:)?(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})( port \d+)?( ssh2)?$""")

TIME_SPEC_REGEX = re.compile(r"""(?P<units>\d*)\s*(?P<period>[smhdwy])?""")

ALLOWED_REGEX = re.compile(r"""(?P<first_3bits>\d{1,3}\.\d{1,3}\.\d{1,3}\.)((?P<fourth>\d{1,3})|(?P<ip_wildcard>\*)|\[(?P<ip_range>\d{1,3}\-\d{1,3})\])""")

PREFS_REGEX = re.compile(r"""(?P<name>.*?)\s*[:=]\s*(?P<value>.*)""")

FAILED_DOVECOT_ENTRY_REGEX = re.compile(r"""dovecot.*authentication failure.*ruser=(?P<user>\S+).*rhost=(?P<host>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}).*""")



import logging
import os
from socket import getfqdn, gethostbyname

from .constants import ALLOWED_HOSTS, ALLOWED_WARNED_HOSTS
from .regex import ALLOWED_REGEX
from .util import is_true

logger = logging.getLogger("AllowedHosts")
debug, warn = logger.debug, logger.warn

class AllowedHosts(object):
    def __init__(self, prefs):
        debug("initializing AllowedHosts")
        work_dir = prefs.get("WORK_DIR")
        self.hostname_lookup = is_true(prefs.get("ALLOWED_HOSTS_HOSTNAME_LOOKUP"))
        self.allowed_path = os.path.join(work_dir, ALLOWED_HOSTS)
        self.warned_path = os.path.join(work_dir, ALLOWED_WARNED_HOSTS)
        self.allowed_hosts = {}
        self.warned_hosts = {}
        self.new_warned_hosts = []
        self.load_hosts()
        self.load_warned_hosts()
        debug("done initializing AllowedHosts")

    def __contains__(self, ip_addr):
        if ip_addr in self.allowed_hosts: return 1
        else: return 0

    def dump(self):
        print("Dumping AllowedHosts")
        print(list(self.allowed_hosts.keys()))


    def load_hosts(self):
        try:
            fp = open(self.allowed_path, "r")
        except Exception as e:
            debug("Could not open %s - %s", self.allowed_path, str(e))
            return

        for line in fp:
            line = line.strip()
            if not line or line[0] == '#': continue

            m = ALLOWED_REGEX.match(line)
            debug("line: %s - regex match?   %s", line, m is not None)
            if m:
                # line contains an ip address
                first3 = m.group('first_3bits')
                fourth = m.group('fourth')
                wildcard = m.group('ip_wildcard')
                ip_range = m.group('ip_range')

                if fourth:
                    self.allowed_hosts["%s%s" % (first3, fourth)] = 1
                    self.add_hostname("%s%s" % (first3, fourth))
                elif wildcard:
                    for i in range(256):
                        self.allowed_hosts["%s%s" % (first3, i)] = 1
                        self.add_hostname("%s%s" % (first3, i))
                else:
                    start, end = ip_range.split("-")
                    for i in range(int(start), int(end)):
                        self.allowed_hosts["%s%d" % (first3, i)] = 1
                        self.add_hostname("%s%s" % (first3, i))
            else:
                # assume that line contains hostname
                self.allowed_hosts[line] = 1
                try:
                    # lookup ip address of host
                    ip = gethostbyname(line)
                    self.allowed_hosts[ip] = 1
                except:
                    pass

        fp.close()
        debug("allowed_hosts: %s", list(self.allowed_hosts.keys()))


    def add_hostname(self, ip_addr):
        if not self.hostname_lookup:
            return
        else:
            hostname = getfqdn(ip_addr)
            if hostname != ip_addr:
                self.allowed_hosts[hostname] = 1


    def add_warned_host(self, host):
        #debug("warned_hosts: %s", self.warned_hosts.keys())

        if host not in self.warned_hosts:
            debug("%s not in warned hosts" % host)
            self.new_warned_hosts.append(host)
            self.warned_hosts[host] = None


    def get_new_warned_hosts(self):
        return self.new_warned_hosts


    def load_warned_hosts(self):
        try:
            fp = open(self.warned_path, "r")
            for line in fp:
                self.warned_hosts[line.strip()] = None
            fp.close()
        except IOError:
            warn("Couldn't load warned hosts from %s" % self.warned_path)

    def save_warned_hosts(self):
        if not self.new_warned_hosts: return
        try:
            fp = open(self.warned_path, "a")
            for host in self.new_warned_hosts:
                fp.write("%s\n" % host)
            fp.close()
        except Exception as e:
            print(e)


    def clear_warned_hosts(self):
        self.new_warned_hosts = []



#!/usr/bin/env python
import os
import platform
import sys
import logging
sys.path.insert(0, '/usr/share/denyhosts')

import DenyHosts.python_version

import getopt
from getopt import GetoptError
import traceback

from DenyHosts.util import die, setup_logging, is_true
from DenyHosts.lockfile import LockFile
from DenyHosts.prefs import Prefs
from DenyHosts.version import VERSION
from DenyHosts.deny_hosts import DenyHosts
from DenyHosts.denyfileutil import Purge, PurgeIP, Migrate, UpgradeTo099
from DenyHosts.constants import *
from DenyHosts.sync import Sync

info = logging.getLogger("denyhosts").info
#################################################################################

def usage():
    print("Usage:")
    print("%s [-f logfile | --file=logfile] [ -c configfile | --config=configfile] [-i | --ignore] [-n | --noemail] [--purge] [--purge-all] [--purgeip=ip] [--migrate] [--daemon] [--sync] [--version]" % sys.argv[0])
    print("\n\n")
    print(" --config: The pathname of the configuration file")
    print(" --file:   The name of log file to parse")
    print(" --ignore: Ignore last processed offset (start processing from beginning)")
    print(" --noemail: Do not send an email report")
    print(" --unlock: if lockfile exists, remove it and run as normal")
    print(" --migrate: migrate your HOSTS_DENY file so that it is suitable for --purge")
    print(" --purge: expire entries older than your PURGE_DENY setting")
    print(" --purge-all: expire all entries")
    print(" --purgeip: expire designated IP entry immediately")
    print(" --daemon: run DenyHosts in daemon mode")
    print(" --foreground: run DenyHosts in foreground mode")
    print(" --sync: run DenyHosts synchronization mode")
    print(" --version: Prints the version of DenyHosts and exits")

    print("\n")
    print("Note: multiple --file args can be processed. ")
    print("If multiple files are provided, --ignore is implied")
    print("\n")
    print("Note: multiple --purgeip arguments can be processed. ")
    print("\n")
    print("When run in --daemon mode the following flags are ignored:")
    print("     --file, --purge, --purge-all, --purgeip, --migrate, --sync, --verbose")


#################################################################################




#################################################################################


if __name__ == '__main__':
    logfiles = []
    purgeip_list = []
    config_file = CONFIG_FILE
    ignore_offset = 0
    noemail = 0
    verbose = 0
    migrate = 0
    purge = 0
    purge_all = 0
    sync_mode = 0
    daemon = 0
    foreground = 0
    enable_debug = 0
    purgeip = 0
    upgrade099 = 0
    unlock = 0
    args = sys.argv[1:]
    try:
        (opts, getopts) = getopt.getopt(args, 'f:c:dinuvps?hV',
                                        ["file=", "ignore", "verbose", "debug",
                                         "help", "noemail", "config=", "version",
                                         "migrate", "purge", "purge-all", "purgeip", "daemon", "foreground",
                                         "unlock", "sync", "upgrade099"])
    except GetoptError:
        print("\nInvalid command line option detected.")
        usage()
        sys.exit(1)

    for opt, arg in opts:
        if opt in ('-h', '-?', '--help'):
            usage()
            sys.exit(0)
        if opt in ('-f', '--file'):
            logfiles.append(arg)
        if opt in ('-i', '--ignore'):
            ignore_offset = 1
        if opt in ('-n', '--noemail'):
            noemail = 1
        if opt in ('-v', '--verbose'):
            verbose = 1
        if opt in ('-d', '--debug'):
            enable_debug = 1
        if opt in ('-c', '--config'):
            config_file = arg
        if opt in ('-m', '--migrate'):
            migrate = 1
        if opt in ('-p', '--purge'):
            purge = 1
        if opt in ('-s', '--sync'):
            sync_mode = 1
        if opt in ('-s', '--unlock'):
            unlock = 1
        if opt == '--daemon':
            daemon = 1
        if opt == '--foreground':
            foreground = 1
        if opt == '--purge-all':
            purge_all = 1
        if opt == '--purgeip':
            purgeip_list.append(arg)
            purgeip = 1
        if opt == '--upgrade099':
            upgrade099 = 1
        if opt == '--version':
            print("DenyHosts version:", VERSION)
            sys.exit(0)

    # This is generally expected to be in the environment, but there's no
    # non-hackish way to get systemd to set it, so just hack it in here.
    os.environ['HOSTNAME'] = platform.node()

    prefs = Prefs(config_file)

    first_time = 0
    try:
        if not os.path.exists( prefs.get('WORK_DIR') ):
             os.makedirs(prefs.get('WORK_DIR'))
             first_time = 1
    except Exception as e:
        if e[0] != 17:
            print(e)
            sys.exit(1)

    # On some operating systems the file /etc/hosts.deny (or
    # whatever HOSTS_DENY is set to, may not exist. We should
    # "touch" the file to make sure it is there to avoid errors later.
    try:
        host_filename = prefs.get("HOSTS_DENY")
        if (host_filename): 
            fp = open( prefs.get("HOSTS_DENY"), "a" )
            fp.close();
    except Exception as e:
        print("Unable to create file specified by HOSTS_DENY variable.")

    setup_logging(prefs, enable_debug, verbose, daemon)

    if not logfiles or daemon:
        logfiles = [prefs.get('SECURE_LOG')]
    elif len(logfiles) > 1:
        ignore_offset = 1

    if not prefs.get('ADMIN_EMAIL'): noemail = 1

    lock_file = LockFile(prefs.get('LOCK_FILE'))

    if unlock:
        if os.path.isfile( prefs.get('LOCK_FILE') ):
           lock_file.remove()

    lock_file.create()

    if upgrade099 and not (daemon or foreground):
        if not prefs.get('PURGE_DENY'):
            lock_file.remove()
            die("You have supplied the --upgrade099 flag, however you have not set PURGE_DENY in your configuration file")
        else:
            u = UpgradeTo099(prefs.get("HOSTS_DENY"))

    if migrate and not (daemon or foreground):
        if not prefs.get('PURGE_DENY'):
            lock_file.remove()
            die("You have supplied the --migrate flag however you have not set PURGE_DENY in your configuration file.")
        else:
            m = Migrate(prefs.get("HOSTS_DENY"))

    # clear out specific IP addresses
    if purgeip and not daemon:
        if len(purgeip_list) < 1:
            lock_file.remove()
            die("You have provided the --purgeip flag however you have not listed any IP addresses to purge.")
        else:
            try:
                p = PurgeIP(prefs,
                          purgeip_list)

            except Exception as e:
                lock_file.remove()
                die(str(e))


    # Try to purge old records without any delay
    if purge_all and not daemon:
         purge_time = 1
         try:
            p = Purge(prefs, purge_time)
         except Exception as e:
            lock_file.remove()
            die(str(e))

    if purge and not (daemon or foreground):
        purge_time = prefs.get('PURGE_DENY')
        if not purge_time:
            lock_file.remove()
            die("You have provided the --purge flag however you have not set PURGE_DENY in your configuration file.")
        else:
            try:
                p = Purge(prefs,
                          purge_time)

            except Exception as e:
                lock_file.remove()
                die(str(e))

    try:
        for f in logfiles:
            dh = DenyHosts(f, prefs, lock_file, ignore_offset,
                           first_time, noemail, daemon, foreground)
    except KeyboardInterrupt:
        pass
    except SystemExit as e:
        pass
    except Exception as e:
        traceback.print_exc(file=sys.stdout)
        print("\nDenyHosts exited abnormally")


    if sync_mode and not (daemon or foreground):
        if not prefs.get('SYNC_SERVER'):
            lock_file.remove()
            die("You have provided the --sync flag however your configuration file is missing a value for SYNC_SERVER.")
        sync_upload = is_true(prefs.get("SYNC_UPLOAD"))
        sync_download = is_true(prefs.get("SYNC_DOWNLOAD"))
        if not sync_upload and not sync_download:
           lock_file.remove()
           die("You have provided the --sync flag however your configuration file has SYNC_UPLOAD and SYNC_DOWNLOAD set to false.")
        try:
            sync = Sync(prefs)
            if sync_upload:
                timestamp = sync.send_new_hosts()
            if sync_download:
                new_hosts = sync.receive_new_hosts()
                if new_hosts:
                    # MMR: What is 'info' here?
                    info("received new hosts: %s", str(new_hosts))
                    sync.get_denied_hosts()
                    sync.update_hosts_deny(new_hosts)
                    dh.get_denied_hosts()
                    dh.update_hosts_deny(new_hosts)
            sync.xmlrpc_disconnect()
        except Exception as e:
            lock_file.remove()
            die("Error synchronizing data", e)

    # remove lock file on exit
    lock_file.remove()



from __future__ import print_function, unicode_literals

from os.path import join as ospj
from tempfile import mkdtemp
import time
import unittest

from DenyHosts.counter import CounterRecord
from DenyHosts.prefs import Prefs
from DenyHosts.loginattempt import LoginAttempt
from DenyHosts.constants import (
    ABUSIVE_HOSTS_INVALID,
    ABUSIVE_HOSTS_ROOT,
    ABUSIVE_HOSTS_RESTRICTED,
    ABUSIVE_HOSTS_VALID,
    ABUSED_USERS_INVALID,
    ABUSED_USERS_VALID,
    ABUSED_USERS_AND_HOSTS,
    SUSPICIOUS_LOGINS,
)

# List of 2-tuples:
# [0] filename in which data is stored
# [1] LoginAttempt attribute name that reads from this file
LOGIN_ATTEMPT_ATTRIBUTE_FILENAMES = [
    (ABUSIVE_HOSTS_INVALID, 'get_abusive_hosts_invalid'),
    (ABUSIVE_HOSTS_ROOT, 'get_abusive_hosts_root'),
    (ABUSIVE_HOSTS_RESTRICTED, 'get_abusive_hosts_restricted'),
    (ABUSIVE_HOSTS_VALID, 'get_abusive_hosts_valid'),
    (ABUSED_USERS_INVALID, 'get_abused_users_invalid'),
    (ABUSED_USERS_VALID, 'get_abused_users_valid'),
    (ABUSED_USERS_AND_HOSTS, 'get_abused_users_and_hosts'),
    (SUSPICIOUS_LOGINS, 'get_suspicious_logins'),
]

class LoginAttemptTestBase(unittest.TestCase):
    def setUp(self):
        self.allowed_hosts = ['host1', 'host2']
        self.prefs = Prefs()
        self.work_dir = mkdtemp()
        self.prefs._Prefs__data['WORK_DIR'] = self.work_dir

        keys = [
            'DENY_THRESHOLD_INVALID',
            'DENY_THRESHOLD_VALID',
            'DENY_THRESHOLD_ROOT',
            'DENY_THRESHOLD_RESTRICTED',
        ]
        for key in keys:
            self.prefs._Prefs__data[key] = 0

class BasicLoginAttemptTest(LoginAttemptTestBase):
    def test_no_data_file(self):
        login_attempt = LoginAttempt(self.prefs, set(self.allowed_hosts))
        for filename, method_name in LOGIN_ATTEMPT_ATTRIBUTE_FILENAMES:
            self.assertFalse(getattr(login_attempt, method_name)())

class LoginAttemptDataFileTest(LoginAttemptTestBase):
    def test_data_files(self):
        login_attempt = LoginAttempt(self.prefs, set(self.allowed_hosts))
        host = 'host'
        count = 1
        asctime = time.asctime()
        test_counter_record = CounterRecord(count=count, date=asctime)
        for filename, method_name in LOGIN_ATTEMPT_ATTRIBUTE_FILENAMES:
            path = ospj(self.work_dir, filename)
            with open(path, 'w') as f:
                print('%s:%d:%s' % (host, count, asctime), file=f)
            data = getattr(login_attempt, method_name)()
            self.assertTrue(host in data)
            # TODO fix this after defining CounterRecord.__eq__
            real_counter_record = data[host]
            self.assertEqual(test_counter_record.get_count(), real_counter_record.get_count())
            self.assertEqual(test_counter_record.get_date(), real_counter_record.get_date())



from __future__ import print_function, unicode_literals

from datetime import datetime, timedelta
import time
import unittest

from DenyHosts.counter import Counter, CounterRecord

class CounterRecordTest(unittest.TestCase):
    def test_init(self):
        c = CounterRecord()
        self.assertEqual(c.get_count(), 0)
        # Counter.__date is initialized with time.asctime(), so there isn't
        # much to test beyond the type
        self.assertTrue(isinstance(c.get_date(), str))

    def test_init_provided_date(self):
        """
        CounterRecord.__date is intended to be a string (for some reason; a datetime
        object would be more appropriate), but any object can be used. Verify that
        what we pass to the constructor is accessible.
        """
        date = object()
        c = CounterRecord(date=date)
        self.assertTrue(c.get_date() is date)

    def test_init_provided_count(self):
        """
        CounterRecord.__count is intended to be numeric, but any object can be used.
        Verify that what we pass to the constructor is accessible.
        """
        count = object()
        c = CounterRecord(count=count)
        self.assertTrue(c.get_count() is count)

    def test_str(self):
        """
        CounterRecord.__str__ is actually used in PurgeCounter.write_data, so it's
        worth testing
        """
        count = 1
        date = object()
        c = CounterRecord(count=count, date=date)
        string = '%d:%s' % (count, date)
        self.assertEqual(str(c), string)

    def test_add(self):
        """
        CounterRecord.__add__ is *horrible* design, but that's how it's been for a
        very long time. I want test coverage for the current behavior before making
        any changes.
        """
        c = CounterRecord()
        orig_date = c.get_date()
        self.assertEqual(c.get_count(), 0)
        increment = 4
        # !
        c + increment
        self.assertEqual(c.get_count(), increment)
        # Original attempt: self.assertNotEqual(c.get_date(), orig_date)
        # time.asctime only provides seconds in that string representation of the
        # date, though, so just verify that the two strings are different objects
        # since they'll usually be equal
        self.assertFalse(c.get_date() is orig_date)

    def test_reset_count(self):
        c = CounterRecord()
        c + 1
        orig_date = c.get_date()
        c.reset_count()
        self.assertEqual(c.get_count(), 0)
        self.assertTrue(c.get_date() is orig_date)

    def test_age_count_newer(self):
        """
        Initialize a CounterRecord to one hour ago, then call age_count with 2 hours
        to verify that the count won't reset. ("Reset if the stored date is older than
        2 hours ago")
        """
        one_hour_ago = datetime.now() - timedelta(hours=1)
        date_str = time.asctime(one_hour_ago.timetuple())
        count = object()
        c = CounterRecord(count=count, date=date_str)
        c.age_count(2 * 60 * 60)
        self.assertEqual(c.get_count(), count)

    def test_age_count_older(self):
        """
        Initialize a CounterRecord to one hour ago, then reset the count by passing 0
        to age_count (i.e. "reset if the stored date is older than now")
        """
        one_hour_ago = datetime.now() - timedelta(hours=1)
        date_str = time.asctime(one_hour_ago.timetuple())
        count = object()
        c = CounterRecord(count=count, date=date_str)
        c.age_count(0)
        self.assertEqual(c.get_count(), 0)

class CounterTest(unittest.TestCase):
    def test_init(self):
        c = Counter()
        self.assertEqual(len(c), 0)

    def test_missing_key(self):
        c = Counter()
        key = 'key'
        value = c[key]
        self.assertEqual(value.get_count(), 0)
        self.assertTrue(key in c)

    def test_existing_key(self):
        key = 'key'
        value = object()
        c = Counter()
        c[key] = value
        self.assertTrue(c[key] is value)



from __future__ import print_function, unicode_literals

from os.path import dirname, join as ospj
from tempfile import mkdtemp
import unittest

from DenyHosts.constants import PURGE_HISTORY
from DenyHosts.counter import Counter
from DenyHosts.prefs import Prefs
from DenyHosts.purgecounter import PurgeCounter

class EmptyPurgeCounterTest(unittest.TestCase):
    """
    Tests creating a PurgeCounter object with no data file. This should:
    1) not throw an exception, and
    2) result in empty sets of hosts.

    We ensure that there's no data file by making an empty temporary
    directory for this test, and assinging that path to the appropriate
    Prefs key.
    """
    def setUp(self):
        self.prefs = Prefs()
        self.prefs._Prefs__data['WORK_DIR'] = mkdtemp()

    def test_no_data_file(self):
        purgecounter = PurgeCounter(self.prefs)
        self.assertFalse(purgecounter.get_banned_for_life())
        self.assertFalse(purgecounter.get_data())

class PurgeCounterTestThreshold0(unittest.TestCase):
    """
    Tests creating a Restricted object with a data file.
    """
    def setUp(self):
        self.prefs = Prefs()
        self.prefs._Prefs__data['WORK_DIR'] = ospj(dirname(__file__), 'data/purgecounter/static')

    def test_data_file(self):
        purgecounter = PurgeCounter(self.prefs)
        self.assertFalse(purgecounter.get_banned_for_life())
        self.assertEqual(len(purgecounter.get_data()), 1)

class PurgeCounterTestThreshold1(unittest.TestCase):
    """
    Tests creating a Restricted object with a data file.
    """
    def setUp(self):
        self.prefs = Prefs()
        self.prefs._Prefs__data['WORK_DIR'] = ospj(dirname(__file__), 'data/purgecounter/static')
        self.prefs._Prefs__data['PURGE_THRESHOLD'] = 1

    def test_data_file(self):
        purgecounter = PurgeCounter(self.prefs)
        self.assertEqual(len(purgecounter.get_banned_for_life()), 1)
        self.assertEqual(len(purgecounter.get_data()), 1)

class PurgeCounterTestWriteData(unittest.TestCase):
    """
    Tests creating a Restricted object with a data file.
    """
    def setUp(self):
        self.prefs = Prefs()
        work_dir = ospj(dirname(__file__), 'data/purgecounter/dynamic')
        self.filename = ospj(work_dir, PURGE_HISTORY)
        self.prefs._Prefs__data['WORK_DIR'] = work_dir
        self.counter = Counter()
        host = 'host'
        count = 1
        self.counter[host] += count
        self.test_string = '%s:%s\n' % (host, self.counter[host])

    def test_data_file(self):
        purgecounter = PurgeCounter(self.prefs)
        purgecounter.write_data(self.counter)

        with open(self.filename) as f:
            self.assertEqual(f.read(), self.test_string)

class PurgeCounterTestIncrement(unittest.TestCase):
    def setUp(self):
        self.prefs = Prefs()
        work_dir = mkdtemp()
        self.filename = ospj(work_dir, PURGE_HISTORY)
        self.prefs._Prefs__data['WORK_DIR'] = work_dir
        self.hosts = set(['host1', 'host2'])

    def test_increment(self):
        purge_counter = PurgeCounter(self.prefs)
        self.assertFalse(purge_counter.get_data())
        purge_counter.increment(self.hosts)
        data = purge_counter.get_data()
        self.assertEqual(set(data), self.hosts)



import os
import shutil
import time
import logging

from .constants import DENY_DELIMITER, ENTRY_DELIMITER
from .loginattempt import AbusiveHosts
from .util import parse_host
from . import plugin
from .purgecounter import PurgeCounter

debug = logging.getLogger("denyfileutil").debug
info = logging.getLogger("denyfileutil").info
warn = logging.getLogger("denyfileutil").warn

class DenyFileUtilBase(object):
    def __init__(self, deny_file, extra_file_id=""):
        self.deny_file = deny_file
        self.backup_file = "%s.%s.bak" % (deny_file, extra_file_id)
        self.temp_file = "%s.%s.tmp" % (deny_file, extra_file_id)

    def backup(self):
        try:
            shutil.copy(self.deny_file, self.backup_file)
        except Exception as e:
            warn(str(e))

    def replace(self):
        # overwrites deny_file with contents of temp_file
        try:
            os.rename(self.temp_file, self.deny_file)
        except Exception as e:
            print(e)

    def remove_temp(self):
        try:
            os.unlink(self.temp_file)
        except OSError:
            pass

    def create_temp(self, data_list):
        raise Exception("Not Imlemented")


    def get_data(self):
        data = []
        try:
            fp = open(self.backup_file, "r")
            data = fp.readlines()
            fp.close()
        except IOError:
            pass
        return data

#################################################################################

class Migrate(DenyFileUtilBase):
    def __init__(self, deny_file):
        print("")
        print("**** WARNING ****")
        print("migrate switch will migrate ALL your entries in your HOSTS_DENY file")
        print("and this can be potentially dangerous, if you have some entry that ")
        print("you won't purge")
        print("")
        print("If you don't understand, please type 'No' and")
        print("read /usr/share/doc/denyhosts/README.Debian")
        print("for more info")
        print("")
        response = input("Are you sure that you want do this? (Yes/No)")
        if response == "Yes":
            DenyFileUtilBase.__init__(self, deny_file, "migrate")
            self.backup()
            self.create_temp(self.get_data())
            self.replace()
        else:
            print("nothing done")

    def create_temp(self, data):
        try:
            fp = open(self.temp_file, "w")
            os.chmod(self.temp_file, 0o644)
            for line in data:
                if line.find("#") != -1:
                    fp.write(line)
                    continue

                line = line.strip()
                if not line:
                    fp.write("\n")
                    continue

                fp.write("%s %s%s%s\n" % (DENY_DELIMITER,
                                          time.asctime(),
                                          ENTRY_DELIMITER,
                                          line))
                fp.write("%s\n" % line)

            fp.close()
        except Exception as e:
            raise e



#################################################################################

class UpgradeTo099(DenyFileUtilBase):
    def __init__(self, deny_file):
        DenyFileUtilBase.__init__(self, deny_file, "0.9.9")
        self.backup()
        self.create_temp(self.get_data())
        self.replace()

    def create_temp(self, data):
        try:
            fp = open(self.temp_file, "w")
            for line in data:
                if line.find("#") == 0:
                    fp.write(line)
                    continue

                line = line.strip()
                if not line:
                    fp.write("\n")
                    continue

                delimiter_idx = line.find(DENY_DELIMITER)

                if delimiter_idx == -1:
                    fp.write("%s\n" % line)
                    continue

                entry = line[:delimiter_idx].strip()
                fp.write("%s%s%s\n" % (line[delimiter_idx:],
                                       ENTRY_DELIMITER,
                                       entry))
                fp.write("%s\n" % entry)
            fp.close()
        except Exception as e:
            raise e

#################################################################################

class Purge(DenyFileUtilBase):
    def __init__(self, prefs, cutoff):
        deny_file = prefs.get('HOSTS_DENY')
        DenyFileUtilBase.__init__(self, deny_file, "purge")
        work_dir = prefs.get('WORK_DIR')
        self.purge_threshold = prefs['PURGE_THRESHOLD']
        self.purge_counter = PurgeCounter(prefs)

        self.cutoff = int(time.time()) - cutoff
        debug("relative cutoff: %ld (seconds)", cutoff)
        debug("absolute cutoff: %ld (epoch)", self.cutoff)
        info("purging entries older than: %s",
             time.asctime(time.localtime(self.cutoff)))

        self.backup()

        purged_hosts = self.create_temp(self.get_data())
        num_purged = len(purged_hosts)
        if num_purged > 0:
            self.replace()
            abusive_hosts = AbusiveHosts(prefs)
            abusive_hosts.purge_hosts(purged_hosts)
            abusive_hosts.save_abusive_hosts()
            self.purge_counter.increment(purged_hosts)
        else:
            self.remove_temp()

        info("num entries purged: %d", num_purged)
        plugin_purge = prefs.get('PLUGIN_PURGE')
        if plugin_purge:
            plugin.execute(plugin_purge, purged_hosts)


    def create_temp(self, data):
        purged_hosts = []
        banned = self.purge_counter.get_banned_for_life()

        try:
            fp = open(self.temp_file, "w")
            os.chmod(self.temp_file, 0o644)
            offset = 0
            num_lines = len(data)
            while offset < num_lines:
                line = data[offset]
                offset += 1
                if not line.startswith(DENY_DELIMITER):
                    fp.write(line)
                    continue
                else:
                    if offset == num_lines:
                        warn("DenyHosts comment line at end of file")
                        fp.write(line)
                        continue

                    timestamp = None
                    try:
                        rest = line.lstrip(DENY_DELIMITER)
                        timestamp, host_verify = rest.split(ENTRY_DELIMITER)
                        tm = time.strptime(timestamp)
                    except Exception as e:
                        warn("Parse error -- Ignorning timestamp: %s for: %s", timestamp, line)
                        warn("exception: %s", str(e))
                        # ignoring bad time string
                        fp.write(line)
                        continue

                    epoch = int(time.mktime(tm))
                    #print entry, epoch, self.cutoff

                    if self.cutoff > epoch:
                        # this entry should be purged
                        entry = data[offset]
                        if host_verify != entry:
                            warn("%s purge verification failed: %s vs. %s",
                                 self.deny_file,
                                 host_verify.rstrip(),
                                 entry.rstrip())

                            fp.write(line)
                            continue
                        host = parse_host(entry)
                        if host and host not in banned:
                            # purge
                            purged_hosts.append(host)

                            # increment offset past purged line
                            offset += 1
                        continue
                    else:
                        fp.write(line)
                        continue

            fp.close()
        except Exception as e:
            raise e
        return purged_hosts

#################################################################################

class PurgeIP(DenyFileUtilBase):
    def __init__(self, prefs, purgeip_list):
        deny_file = prefs.get('HOSTS_DENY')
        DenyFileUtilBase.__init__(self, deny_file, "purgeip")
        work_dir = prefs.get('WORK_DIR')
        self.purge_counter = PurgeCounter(prefs)

        info("purging listed IP addresses.",)

        self.backup()

        purged_hosts = purgeip_list
        num_purged = len(purged_hosts)
        if num_purged > 0:
            self.replace()
            abusive_hosts = AbusiveHosts(prefs)
            abusive_hosts.purge_hosts(purged_hosts)
            abusive_hosts.save_abusive_hosts()
            self.purge_counter.increment(purged_hosts)
        else:
            self.remove_temp()

        info("num entries purged: %d", num_purged)
        plugin_purge = prefs.get('PLUGIN_PURGE')
        if plugin_purge:
            plugin.execute(plugin_purge, purged_hosts)



from __future__ import print_function, unicode_literals

import unittest

from DenyHosts.regex import FAILED_ENTRY_REGEX_MAP, SSHD_FORMAT_REGEX

# From https://bugs.gentoo.org/show_bug.cgi?id=157163
CVE_2006_6301_LINE = ' sshd[23964]: Invalid user foo from 123.123.123.123 from 127.0.0.1'
CVE_2006_6301_USER = 'foo from 123.123.123.123'
CVE_2006_6301_HOST = '127.0.0.1'

# From https://bugzilla.redhat.com/show_bug.cgi?id=237449
CVE_2007_5715_LINE = ' sshd[29961]: User root from 122.36.2.10 not allowed because not listed in AllowUsers'
CVE_2007_5715_USER = 'root'
CVE_2007_5715_HOST = '122.36.2.10'

# From Helmut Grohne's disclosure of this CVE
CVE_2013_6890_LINES = [
    ' sshd[123]: Invalid user Invalid user root from 123.123.123.123 from 21.21.21.21',
    ' sshd[123]: input_userauth_request: invalid user Invalid user root from 123.123.123.123 [preauth]',
    ' sshd[123]: Connection closed by 21.21.21.21 [preauth]',
]

CVE_2013_6890_USER = 'Invalid user root from 123.123.123.123'
CVE_2013_6890_HOST = '21.21.21.21'

# TODO separate matching behavior into a common function
# that's also used by the DenyHosts daemon class, instead of
# doing it ourselves in every test case here

class CVEsTestCase(unittest.TestCase):
    def test_cve_2006_6301(self):
        user = None
        host = None
        sshd_m = SSHD_FORMAT_REGEX.match(CVE_2006_6301_LINE)
        if sshd_m:
            message = sshd_m.group('message')

            for rx in FAILED_ENTRY_REGEX_MAP.values():
                m = rx.search(message)
                if m:
                    user = m.group('user')
                    host = m.group('host')
        self.assertEqual(user, CVE_2006_6301_USER)
        self.assertEqual(host, CVE_2006_6301_HOST)

    def test_cve_2007_4323(self):
        pass

    def test_cve_2007_5715(self):
        user = None
        host = None
        sshd_m = SSHD_FORMAT_REGEX.match(CVE_2007_5715_LINE)
        if sshd_m:
            message = sshd_m.group('message')

            for rx in FAILED_ENTRY_REGEX_MAP.values():
                m = rx.search(message)
                if m:
                    user = m.group('user')
                    host = m.group('host')
        self.assertEqual(user, CVE_2007_5715_USER)
        self.assertEqual(host, CVE_2007_5715_HOST)

    def test_cve_2013_6890(self):
        user = None
        host = None
        # There's no harm in iterating over all three lines even though
        # the first contains what we want. The second and third lines
        # don't match any of the 'failed entry' regexes.
        for line in CVE_2013_6890_LINES:
            sshd_m = SSHD_FORMAT_REGEX.match(line)
            if sshd_m:
                message = sshd_m.group('message')

                for rx in FAILED_ENTRY_REGEX_MAP.values():
                    m = rx.search(message)
                    if m:
                        user = m.group('user')
                        host = m.group('host')
        self.assertEqual(user, CVE_2013_6890_USER)
        self.assertEqual(host, CVE_2013_6890_HOST)



import os
import logging
import errno
from .util import is_true

from .counter import Counter, CounterRecord
from .constants import *

debug = logging.getLogger("loginattempt").debug
info = logging.getLogger("loginattempt").info

class LoginAttempt(object):
    def __init__(self, prefs, allowed_hosts, suspicious_always=1,
                 first_time=0, fetch_all=1, restricted=None):
        if restricted is None:
            restricted = set()
        self.__restricted = restricted

        self.__work_dir = prefs.get('WORK_DIR')

        self.__deny_threshold_invalid = prefs.get('DENY_THRESHOLD_INVALID')
        self.__deny_threshold_valid = prefs.get('DENY_THRESHOLD_VALID')
        self.__deny_threshold_root = prefs.get('DENY_THRESHOLD_ROOT')
        self.__deny_threshold_restricted = prefs.get('DENY_THRESHOLD_RESTRICTED')

        self.__age_reset_invalid = prefs.get('AGE_RESET_INVALID')
        self.__age_reset_valid = prefs.get('AGE_RESET_VALID')
        self.__age_reset_root = prefs.get('AGE_RESET_ROOT')
        self.__age_reset_restricted = prefs.get('AGE_RESET_RESTRICTED')

        self.__reset_on_success = is_true(prefs.get('RESET_ON_SUCCESS'))

        self.__first_time = first_time
        self.__suspicious_always = suspicious_always
        self.__allowed_hosts = allowed_hosts

        if fetch_all:
            self.__suspicious_logins = self.get_suspicious_logins()
            self.__valid_users = self.get_abused_users_valid()
            self.__invalid_users = self.get_abused_users_invalid()
            self.__valid_users_and_hosts = self.get_abused_users_and_hosts()
            self.__abusive_hosts_valid = self.get_abusive_hosts_valid()
            self.__abusive_hosts_invalid = self.get_abusive_hosts_invalid()
            self.__abusive_hosts_root = self.get_abusive_hosts_root()
            self.__abusive_hosts_restricted = self.get_abusive_hosts_restricted()
            self.__new_suspicious_logins = Counter()


    def get_new_suspicious_logins(self):
        return self.__new_suspicious_logins


    def add(self, user, host, success, invalid):
        user_host_key = "%s - %s" % (user, host)

        if host:
            if self.__age_reset_invalid:
                self.__abusive_hosts_invalid[host].age_count(self.__age_reset_invalid)
            if self.__age_reset_valid:
                self.__abusive_hosts_valid[host].age_count(self.__age_reset_valid)
            if self.__age_reset_restricted:
                self.__abusive_hosts_restricted[host].age_count(self.__age_reset_restricted)
            if self.__age_reset_root:
                self.__abusive_hosts_root[host].age_count(self.__age_reset_root)


        if success and self.__reset_on_success:
            info("resetting count for: %s", host)
            self.__abusive_hosts_valid[host].reset_count()
            # ??? maybe:
            self.__abusive_hosts_invalid[host].reset_count()


        if success and self.__abusive_hosts_invalid[host].get_count() > self.__deny_threshold_invalid: 
            num_failures = self.__valid_users_and_hosts.get(user_host_key, 0)
            self.__suspicious_logins[user_host_key] += 1
            if self.__suspicious_always or host not in self.__allowed_hosts:
                self.__new_suspicious_logins[user_host_key] += 1
        elif not success:
            if user in self.__restricted:
                self.increment_count(host,
                                     self.__abusive_hosts_restricted,
                                     self.__age_reset_restricted)

            if invalid:
                # username is invalid
                self.increment_count(host,
                                     self.__abusive_hosts_invalid,
                                     self.__age_reset_invalid)

                self.__invalid_users[user] += 1
            else:
                # username is valid
                self.increment_count(user,
                                     self.__valid_users)

                self.increment_count(user_host_key,
                                     self.__valid_users_and_hosts)

                if user == 'root':
                    self.increment_count(host,
                                         self.__abusive_hosts_root,
                                         self.__age_reset_root)
                elif user in self.__restricted:
                    self.increment_count(host,
                                         self.__abusive_hosts_restricted,
                                         self.__age_reset_restricted)
                else:
                    self.increment_count(host,
                                         self.__abusive_hosts_valid,
                                         self.__age_reset_valid)

    def increment_count(self, key, count_inst, age_reset=None):
        #if not count_inst.has_key(key) or count_inst.has_key(key) and count_inst[key] is None:
        #    count_inst[key] = CounterRecord(0)

        #debug(count_inst)
        if age_reset:
            count_inst[key].age_count(age_reset)
        #debug(count_inst)
        count_inst[key] += 1
        #debug(count_inst)


    def get_abusive_hosts_invalid(self):
        return self.__get_stats(ABUSIVE_HOSTS_INVALID)

    def get_abusive_hosts_root(self):
        return self.__get_stats(ABUSIVE_HOSTS_ROOT)

    def get_abusive_hosts_restricted(self):
        return self.__get_stats(ABUSIVE_HOSTS_RESTRICTED)

    def get_abusive_hosts_valid(self):
        return self.__get_stats(ABUSIVE_HOSTS_VALID)

    def get_abused_users_invalid(self):
        return self.__get_stats(ABUSED_USERS_INVALID)

    def get_abused_users_valid(self):
        return self.__get_stats(ABUSED_USERS_VALID)

    def get_abused_users_and_hosts(self):
        return self.__get_stats(ABUSED_USERS_AND_HOSTS)

    def get_suspicious_logins(self):
        return self.__get_stats(SUSPICIOUS_LOGINS)

    def __get_stats(self, fname):
        path = os.path.join(self.__work_dir, fname)
        stats = Counter()

        try:
            for line in open(path, "r"):

                try:
                    line = line.strip()
                    parts = line.split(":")
                    name = parts[0]
                    count = parts[1]
                    try:
                        date = ':'.join(parts[2:])
                    except Exception:
                        date = None

                    stats[name] = CounterRecord(int(count), date)
                    #debug("stats[%s] = %s", name, stats[name])
                except Exception as e:
                    ##debug(e)
                    pass
        except IOError as e:
            if e.errno == errno.ENOENT: debug("%s does not exist", fname)
            else: print(e)
        except Exception as e:
            if not self.__first_time: print(e)

        return stats


    def save_all_stats(self):
        self.save_abusive_hosts_valid()
        self.save_abusive_hosts_invalid()
        self.save_abusive_hosts_root()
        self.save_abusive_hosts_restricted()
        self.save_abused_users_valid()
        self.save_abused_users_invalid()
        self.save_abused_users_and_hosts()
        self.save_suspicious_logins()

    def save_abusive_hosts_invalid(self, abusive_hosts=None):
        if abusive_hosts is None:
            abusive_hosts = self.__abusive_hosts_invalid
        self.__save_stats(ABUSIVE_HOSTS_INVALID, abusive_hosts)

    def save_abusive_hosts_root(self, abusive_hosts=None):
        if abusive_hosts is None:
            abusive_hosts = self.__abusive_hosts_root
        self.__save_stats(ABUSIVE_HOSTS_ROOT, abusive_hosts)

    def save_abusive_hosts_restricted(self, abusive_hosts=None):
        if abusive_hosts is None:
            abusive_hosts = self.__abusive_hosts_restricted
        self.__save_stats(ABUSIVE_HOSTS_RESTRICTED, abusive_hosts)

    def save_abusive_hosts_valid(self, abusive_hosts=None):
        if abusive_hosts is None:
            abusive_hosts = self.__abusive_hosts_valid
        self.__save_stats(ABUSIVE_HOSTS_VALID, abusive_hosts)

    def save_abused_users_invalid(self):
        self.__save_stats(ABUSED_USERS_INVALID, self.__invalid_users)

    def save_abused_users_valid(self):
        self.__save_stats(ABUSED_USERS_VALID, self.__valid_users)

    def save_abused_users_and_hosts(self):
        self.__save_stats(ABUSED_USERS_AND_HOSTS, self.__valid_users_and_hosts)

    def save_suspicious_logins(self):
        self.__save_stats(SUSPICIOUS_LOGINS, self.__suspicious_logins)

    def get_deny_hosts(self):
        invalid_hosts = [host for host,count_rec in list(self.__abusive_hosts_invalid.items())
                         if count_rec.get_count() > self.__deny_threshold_invalid]

        root_hosts = [host for host,count_rec in list(self.__abusive_hosts_root.items())
                      if count_rec.get_count() > self.__deny_threshold_root]

        restricted_hosts = [host for host,count_rec in list(self.__abusive_hosts_restricted.items())
                            if count_rec.get_count() > self.__deny_threshold_restricted]

        valid_hosts = [host for host,count_rec in list(self.__abusive_hosts_valid.items())
                       if count_rec.get_count() > self.__deny_threshold_valid]

        deny_set = set(invalid_hosts + valid_hosts + root_hosts + restricted_hosts)
        return list(deny_set)


    def __save_stats(self, fname, stats):
        path = os.path.join(self.__work_dir, fname)
        if stats is None:
            #debug("%s: is none", fname)
            return

        try:
            fp = open(path, "w")
        except Exception as e:
            print(e)
            return

        if not stats:
            # if stats dict is empty-- no data to process
            fp.close()
            return

        keys = list(stats.keys())
        keys.sort()

        for key in keys:
            #debug("")
            #debug("key: %s - stats[key]: %s", key, stats[key])
            #debug("stats: %s", stats)
            #debug("")
            fp.write("%s:%s\n" % (key, stats[key]))
        fp.close()


class AbusiveHosts(LoginAttempt):
    def __init__(self, prefs):
        LoginAttempt.__init__(self,
                              prefs,
                              None,
                              fetch_all = 0)
        self.__abusive_hosts_invalid = self.get_abusive_hosts_invalid()
        self.__abusive_hosts_root = self.get_abusive_hosts_root()
        self.__abusive_hosts_restricted = self.get_abusive_hosts_restricted()
        self.__abusive_hosts_valid = self.get_abusive_hosts_valid()

    def save_abusive_hosts(self):
        LoginAttempt.save_abusive_hosts_invalid(self,
                                                self.__abusive_hosts_invalid)

        LoginAttempt.save_abusive_hosts_root(self,
                                             self.__abusive_hosts_root)

        LoginAttempt.save_abusive_hosts_restricted(self,
                                                   self.__abusive_hosts_restricted)

        LoginAttempt.save_abusive_hosts_valid(self,
                                              self.__abusive_hosts_valid)

    def purge_host(self, host):
        try:
            self.__abusive_hosts_invalid[host] = None
            self.__abusive_hosts_root[host] = None
            self.__abusive_hosts_restricted[host] = None
            self.__abusive_hosts_valid[host] = None
            del self.__abusive_hosts_invalid[host]
            del self.__abusive_hosts_root[host]
            del self.__abusive_hosts_restricted[host]
            del self.__abusive_hosts_valid[host]
        except Exception:
            pass

    def purge_hosts(self, hosts):
        info("purging_hosts: %s", str(hosts))
        for host in hosts:
            self.purge_host(host)



import logging
import os
import re

from .regex import PREFS_REGEX
from .util import die, calculate_seconds

debug = logging.getLogger("prefs").debug
info = logging.getLogger("prefs").info

ENVIRON_REGEX = re.compile(r"""\$\[(?P<environ>[A-Z_]*)\]""")

class Prefs(dict):
    def __getitem__(self, k):
        return self.get(k)

    def __init__(self, path=None, **kwargs):
        super(Prefs, self).__init__(**kwargs)
        # default values for some of the configurable items
        self.__data = {'ADMIN_EMAIL': None,
                       'SUSPICIOUS_LOGIN_REPORT_ALLOWED_HOSTS': 'yes',
                       'HOSTNAME_LOOKUP': 'yes',
                       'SYSLOG_REPORT': 'no',
                       'DAEMON_LOG': '/var/log/denyhosts',
                       'DAEMON_SLEEP': '30s',
                       'DAEMON_PURGE': '1h',
                       'DAEMON_LOG_TIME_FORMAT': None,
                       'DAEMON_LOG_MESSAGE_FORMAT': '%(asctime)s - %(name)-12s: %(levelname)-8s %(message)s',
                       'AGE_RESET_INVALID': None,
                       'AGE_RESET_VALID': None,
                       'AGE_RESET_ROOT': None,
                       'AGE_RESET_RESTRICTED': None,
                       'RESET_ON_SUCCESS': 'no',
                       'PLUGIN_DENY': None,
                       'PLUGIN_PURGE': None,
                       'IPTABLES': None,
                       'BLOCKPORT': None,
                       'PFCTL_PATH': None,
                       'PF_TABLE': None,
                       'PF_TABLE_FILE': None,
                       'SMTP_USERNAME': None,
                       'SMTP_PASSWORD': None,
                       'SMTP_DATE_FORMAT': "%a, %d %b %Y %H:%M:%S %z",
                       'SSHD_FORMAT_REGEX': None,
                       'FAILED_ENTRY_REGEX': None,
                       'FAILED_ENTRY_REGEX2': None,
                       'FAILED_ENTRY_REGEX3': None,
                       'FAILED_ENTRY_REGEX4': None,
                       'FAILED_ENTRY_REGEX5': None,
                       'FAILED_ENTRY_REGEX6': None,
                       'FAILED_ENTRY_REGEX7': None,
                       'FAILED_ENTRY_REGEX8': None,
                       'FAILED_ENTRY_REGEX9': None,
#                       'FAILED_ENTRY_REGEX10': None,
                       'FAILED_DOVECOT_ENTRY_REGEX': None,
                       'DETECT_DOVECOT_LOGIN_ATTEMPTS': "no",
                       'USERDEF_FAILED_ENTRY_REGEX': [],
                       'SUCCESSFUL_ENTRY_REGEX': None,
                       'SYNC_INTERVAL': '1h',
                       'SYNC_SERVER': None,
                       'SYNC_UPLOAD': "yes",
                       'SYNC_DOWNLOAD': "yes",
                       'SYNC_DOWNLOAD_THRESHOLD': 3,
                       'SYNC_DOWNLOAD_RESILIENCY': '5h',
                       'PURGE_THRESHOLD': 0,
                       'ALLOWED_HOSTS_HOSTNAME_LOOKUP': 'no',
                       'ETC_DIR': '/etc'}

        # reqd[0]: required field name
        # reqd[1]: is value required? (False = value can be blank)
        self.reqd = (('DENY_THRESHOLD_INVALID', True),
                     ('DENY_THRESHOLD_VALID', True),
                     ('DENY_THRESHOLD_ROOT', True),
                     ('DENY_THRESHOLD_RESTRICTED', True),
                     ('SECURE_LOG', True),
                     ('LOCK_FILE', True),
                     ('BLOCK_SERVICE', False),
                     ('PURGE_DENY', False),
                     ('HOSTS_DENY', True),
                     ('WORK_DIR', True),
                     ('ETC_DIR', False),
                     ('IPTABLES', False),
                     ('BLOCKPORT', False),
                     ('PFCTL_PATH', False),
                     ('PF_TABLE', False),
                     ('PF_TABLE_FILE', False))

        # the paths for these keys will be converted to
        # absolute pathnames (in the event they are relative)
        # since the --daemon mode requires absolute pathnames
        self.make_abs = ('WORK_DIR',
                         'ETC_DIR',
                         'LOCK_FILE',
                         'SECURE_LOG',
                         'HOSTS_DENY',
                         'DAEMON_LOG',
                         'IPTABLES',
                         'PFCTL_PATH',
                         'PF_TABLE_FILE')

        # these settings are converted to numeric values
        self.to_int = set(('DENY_THRESHOLD',
                           'DENY_THRESHOLD_INVALID',
                           'DENY_THRESHOLD_VALID',
                           'DENY_THRESHOLD_ROOT',
                           'DENY_THRESHOLD_RESTRICTED',
                           'SYNC_DOWNLOAD_THRESHOLD',
                           'PURGE_THRESHOLD'))

        # these settings are converted from timespec format
        # to number of seconds (ie. "1m" -> 60)
        self.to_seconds = set(('PURGE_DENY',
                               'DAEMON_PURGE',
                               'DAEMON_SLEEP',
                               'AGE_RESET_VALID',
                               'AGE_RESET_INVALID',
                               'AGE_RESET_RESTRICTED',
                               'SYNC_INTERVAL',
                               'SYNC_DOWNLOAD_RESILIENCY',
                               'AGE_RESET_ROOT'))


        self.process_defaults()
        if path: self.load_settings(path)

    def process_defaults(self):
        for name in self.to_seconds:
            try:
                self.__data[name] = calculate_seconds(self.__data[name])
            except Exception:
                pass


    def load_settings(self, path):
        try:
            fp = open(path, "r")
        except Exception as e :
            die("Error reading file: %s" % path, e)

        for line in fp:
            line = line.strip()
            if not line or line[0] in ('\n', '#'):
                continue
            try:
                m = PREFS_REGEX.search(line)
                if m:
                    name = m.group('name').upper()
                    value = self.environ_sub(m.group('value'))

                    #print name, value
                    if not value: value = None
                    if name in self.to_int:
                        value = int(value)
                    if name in self.to_seconds and value:
                        value = calculate_seconds(value)
                    if name == 'USERDEF_FAILED_ENTRY_REGEX':
                        self.__data['USERDEF_FAILED_ENTRY_REGEX'].append(re.compile(value))
                    else:
                        self.__data[name] = value
            except Exception as e:
                fp.close()
                die("Error processing configuration parameter %s: %s" % (name, e))
        fp.close()
        self.check_required(path)
        self.make_absolute()


    def make_absolute(self):
        for key in self.make_abs:
            val = self.__data[key]
            if val:
                self.__data[key] = os.path.abspath(val)


    def check_required(self, path):
        ok = 1
        for name_reqd, val_reqd in self.reqd:
            if name_reqd not in self.__data:
                print("Missing configuration parameter: %s" % name_reqd)
                if name_reqd == 'DENY_THRESHOLD_INVALID':
                    print("\nNote: The configuration parameter DENY_THRESHOLD has been renamed")
                    print("      DENY_THRESHOLD_INVALID.  Please update your DenyHosts configuration")
                    print("      file to reflect this change.")

                    if 'DENY_THRESHOLD' in self.__data:
                        print("\n*** Using deprecated DENY_THRESHOLD value for DENY_THRESHOLD_INVALID ***")
                        self.__data['DENY_THRESHOLD_INVALID'] = self.__data['DENY_THRESHOLD']
                    else:
                        ok = 0
                elif name_reqd == 'DENY_THRESHOLD_RESTRICTED':
                    print("\nNote: DENY_THRESHOLD_RESTRICTED has not been defined. Setting this")
                    print("value to DENY_THRESHOLD_ROOT")
                    self.__data['DENY_THRESHOLD_RESTRICTED'] = self.__data['DENY_THRESHOLD_ROOT']
                else:
                    ok = 0
            elif val_reqd and not self.__data[name_reqd]:
                print("Missing configuration value for: %s" % name_reqd)
                ok = 0

        if not ok:
            die("You must correct these problems found in: %s" % path)


    def environ_sub(self, value):
        while True:
            environ_match = ENVIRON_REGEX.search(value)
            if not environ_match: return value
            name = environ_match.group("environ")
            env = os.environ.get(name)
            if not env:
                die("Could not find environment variable: %s" % name)
            value = ENVIRON_REGEX.sub(env, value)


    def get(self, name):
        return self.__data[name]


    def dump(self):
        print("Preferences:")
        keys = list(self.__data.keys())
        for key in keys:
            if key == 'USERDEF_FAILED_ENTRY_REGEX':
                for rx in self.__data[key]:
                    print("   %s: [%s]" % (key, rx.pattern))
            else:
                print("   %s: [%s]" % (key, self.__data[key]))


    def dump_to_logger(self):
        keys = list(self.__data.keys())
        keys.sort()
        info("DenyHosts configuration settings:")
        for key in keys:
            if key == 'USERDEF_FAILED_ENTRY_REGEX':
                for rx in self.__data[key]:
                    info("   %s: [%s]" % (key, rx.pattern))
            else:
                info("   %s: [%s]", key, self.__data[key])



import logging
import os
import time
import sys

if sys.version_info < (3, 0): 
   from xmlrpclib import ServerProxy
else:
   from xmlrpc.client import ServerProxy

from .constants import SYNC_TIMESTAMP, SYNC_HOSTS, SYNC_HOSTS_TMP, SYNC_RECEIVED_HOSTS

logger = logging.getLogger("sync")
debug, info, error, exception = logger.debug, logger.info, logger.error, logger.exception

def get_plural(items):
    if len(items) != 1:
        return "s"
    else:
        return ""

class Sync(object):
    def __init__(self, prefs):
        self.__prefs = prefs
        self.__work_dir = prefs.get('WORK_DIR')
        self.__connected = False
        self.__hosts_added = []

    def xmlrpc_connect(self):
        try:
            self.__server = ServerProxy(self.__prefs.get('SYNC_SERVER'))
            self.__connected = True
        except Exception as e:
            error(str(e))
            self.__connected = False
        return self.__connected

    def xmlrpc_disconnect(self):
        if self.__connected:
            try:
                #self.__server.close()
                self.__server = None
            except Exception:
                pass
            self.__connected = False

    def get_sync_timestamp(self):
        try:
            fp = open(os.path.join(self.__work_dir,
                                   SYNC_TIMESTAMP))
            timestamp = fp.readline()
            timestamp = int(timestamp.strip())
            return timestamp
        except Exception as e:
            error(str(e))
            return 0

    def set_sync_timestamp(self, timestamp):
        try:
            fp = open(os.path.join(self.__work_dir,
                                   SYNC_TIMESTAMP), "w")
            fp.write(timestamp)
        except Exception as e:
            error(e)

    def send_new_hosts(self):
        debug("send_new_hosts()")
        self.__hosts_added = []
        try:
            src_file = os.path.join(self.__work_dir, SYNC_HOSTS)
            dest_file = os.path.join(self.__work_dir, SYNC_HOSTS_TMP)
            os.rename(src_file, dest_file)
        except OSError:
            return False

        hosts = []
        fp = open(dest_file, "r")
        for line in fp.readlines():
            hosts.append(line.strip())
        fp.close()

        try:
            self.__send_new_hosts(hosts)
            info("sent %d new host%s", len(hosts), get_plural(hosts))
            self.__hosts_added = hosts
        except Exception:
            os.rename(dest_file, src_file)
            return False

        try:
            os.unlink(dest_file)
        except OSError:
            pass

        return True

    def __send_new_hosts(self, hosts):
        if not self.__connected and not self.xmlrpc_connect():
            error("Could not initiate xmlrpc connection")
            return

        try:
            self.__server.add_hosts(hosts)
        except Exception as e:
            exception(e)

    def receive_new_hosts(self):
        debug("receive_new_hosts()")

        if not self.__connected and not self.xmlrpc_connect():
            error("Could not initiate xmlrpc connection")
            return
        timestamp = self.get_sync_timestamp()

        try:
            data = self.__server.get_new_hosts(timestamp,
                self.__prefs.get("SYNC_DOWNLOAD_THRESHOLD"),
                                               self.__hosts_added,
                self.__prefs.get("SYNC_DOWNLOAD_RESILIENCY"))
            timestamp = data['timestamp']
            self.set_sync_timestamp(timestamp)
            hosts = data['hosts']
            info("received %d new host%s", len(hosts), get_plural(hosts))
            self.__save_received_hosts(hosts, timestamp)
            return hosts
        except Exception as e:
            exception(e)
            return None

    def __save_received_hosts(self, hosts, timestamp):
        try:
            fp = open(os.path.join(self.__work_dir, SYNC_RECEIVED_HOSTS), "a")
        except IOError as e:
            error(e)
            return

        timestr = time.ctime(float(timestamp))
        for host in hosts:
            fp.write("%s:%s\n" % (host, timestr))
        fp.close()



import logging
import gzip
import os
import signal
from stat import ST_SIZE, ST_INO
import time
try:
    import bz2
    HAS_BZ2 = True
except ImportError:
    HAS_BZ2 = False

from .allowedhosts import AllowedHosts
from .constants import *
from .daemon import createDaemon
from .denyfileutil import Purge
from .filetracker import FileTracker
from .loginattempt import LoginAttempt
from . import plugin
from .regex import *
from .report import Report
from .restricted import Restricted
from .sync import Sync
from .util import die, is_true, parse_host, send_email, is_valid_ip_address
from .version import VERSION

debug = logging.getLogger("denyhosts").debug
info = logging.getLogger("denyhosts").info
error = logging.getLogger("denyhosts").error
warning = logging.getLogger("denyhosts").warning

class DenyHosts(object):
    def __init__(self, logfile, prefs, lock_file,
                 ignore_offset=0, first_time=0,
                 noemail=0, daemon=0, foreground=0):
        self.__denied_hosts = {}
        self.__prefs = prefs
        self.__lock_file = lock_file
        self.__first_time = first_time
        self.__noemail = noemail
        self.__report = Report(prefs.get("HOSTNAME_LOOKUP"), is_true(prefs['SYSLOG_REPORT']))
        self.__daemon = daemon
        self.__foreground = foreground
        self.__sync_server = prefs.get('SYNC_SERVER')
        self.__sync_upload = is_true(prefs.get("SYNC_UPLOAD"))
        self.__sync_download = is_true(prefs.get("SYNC_DOWNLOAD"))
        self.__iptables = prefs.get("IPTABLES")
        self.__blockport = prefs.get("BLOCKPORT")
        self.__pfctl = prefs.get("PFCTL_PATH")
        self.__pftable = prefs.get("PF_TABLE")
        self.__pftablefile = prefs.get("PF_TABLE_FILE")

        r = Restricted(prefs)
        self.__restricted = r.get_restricted()
        info("restricted: %s", self.__restricted)
        self.init_regex()

        try:
            self.file_tracker = FileTracker(self.__prefs.get('WORK_DIR'),
                                            logfile)
        except Exception as e:
            self.__lock_file.remove()
            die("Can't read: %s" % logfile, e)

        self.__allowed_hosts = AllowedHosts(self.__prefs)

        if ignore_offset:
            last_offset = 0
        else:
            last_offset = self.file_tracker.get_offset()


        if last_offset is not None:
            self.get_denied_hosts()
            info("Processing log file (%s) from offset (%ld)",
                 logfile,
                 last_offset)
            offset = self.process_log(logfile, last_offset)
            if offset != last_offset:
                self.file_tracker.save_offset(offset)
                last_offset = offset
        elif not daemon:
            info("Log file size has not changed.  Nothing to do.")

        if daemon and not foreground:
            info("launching DenyHosts daemon (version %s)..." % VERSION)

            #logging.getLogger().setLevel(logging.WARN)

            # remove lock file since createDaemon will
            # create a new pid.  A new lock
            # will be created when runDaemon is invoked
            self.__lock_file.remove()

            retCode = createDaemon()
            if retCode == 0:
                self.runDaemon(logfile, last_offset)
            else:
                die("Error creating daemon: %s (%d)" % (retCode[1], retCode[0]))
        elif foreground:
            info("launching DenyHost (version %s)..." % VERSION)
            self.__lock_file.remove()
            self.runDaemon(logfile, last_offset)


    def killDaemon(self, signum, frame):
        debug("Received SIGTERM")
        info("DenyHosts daemon is shutting down")
        # signal handler

        # self.__lock_file.remove()
        # lock will be freed on SIGTERM by denyhosts.py
        # exception handler (SystemExit)
        sys.exit(0)


    def toggleDebug(self, signum, frame):
        level = logging.getLogger().getEffectiveLevel()
        if level == logging.INFO:
            level = logging.DEBUG
            name = "DEBUG"
        else:
            level = logging.INFO
            name = "INFO"
        info("setting debug level to: %s", name)
        logging.getLogger().setLevel(level)


    def runDaemon(self, logfile, last_offset):
        #signal.signal(signal.SIGHUP, self.killDaemon)
        signal.signal(signal.SIGTERM, self.killDaemon)
        signal.signal(signal.SIGUSR1, self.toggleDebug)
        info("DenyHost daemon is now running, pid: %s", os.getpid())
        info("send daemon process a TERM signal to terminate cleanly")
        info("  eg.  kill -TERM %s", os.getpid())
        self.__lock_file.create()

        info("monitoring log: %s", logfile)
        daemon_sleep = self.__prefs.get('DAEMON_SLEEP')
        purge_time = self.__prefs.get('PURGE_DENY')
        sync_time = self.__prefs.get('SYNC_INTERVAL')
        info("sync_time: %s", str(sync_time))

        if purge_time:
            daemon_purge = self.__prefs.get('DAEMON_PURGE')
            daemon_purge = max(daemon_sleep, daemon_purge)
            purge_sleep_ratio = daemon_purge / daemon_sleep
            self.purge_counter = 0
            info("daemon_purge:      %ld", daemon_purge)
            info("daemon_sleep:      %ld", daemon_sleep)
            info("purge_sleep_ratio: %ld", purge_sleep_ratio)
        else:
            purge_sleep_ratio = None
            info("purging of %s is disabled", self.__prefs.get('HOSTS_DENY'))


        if sync_time and self.__sync_server:
            if sync_time < SYNC_MIN_INTERVAL:
                info("SYNC_INTERVAL (%d) should be atleast %d",
                     sync_time,
                     SYNC_MIN_INTERVAL)
                sync_time = SYNC_MIN_INTERVAL
            sync_time = max(daemon_sleep, sync_time)
            info("sync_time:      : %ld", sync_time)
            sync_sleep_ratio = sync_time / daemon_sleep
            self.sync_counter = 0
            info("sync_sleep_ratio: %ld", sync_sleep_ratio)
        else:
            sync_sleep_ratio = None
            info("denyhost synchronization disabled")

        self.daemonLoop(logfile, last_offset, daemon_sleep,
                        purge_time, purge_sleep_ratio, sync_sleep_ratio)


    def daemonLoop(self, logfile, last_offset, daemon_sleep,
                   purge_time, purge_sleep_ratio, sync_sleep_ratio):

        fp = open(logfile, "r")
        inode = os.fstat(fp.fileno())[ST_INO]

        while 1:

            try:
                curr_inode = os.stat(logfile)[ST_INO]
            except OSError:
                info("%s has been deleted", logfile)
                self.sleepAndPurge(daemon_sleep,
                                   purge_time,
                                   purge_sleep_ratio)
                continue

            if curr_inode != inode:
                info("%s has been rotated", logfile)
                inode = curr_inode
                try:
                    fp.close()
                except IOError:
                    pass

                fp = open(logfile, "r")
                # this ultimately forces offset (if not 0) to be < last_offset
                last_offset = sys.maxsize


            offset = os.fstat(fp.fileno())[ST_SIZE]
            if last_offset is None:
                last_offset = offset

            if offset > last_offset:
                # new data added to logfile
                debug("%s has additional data", logfile)

                self.get_denied_hosts()
                last_offset = self.process_log(logfile, last_offset)

                self.file_tracker.save_offset(last_offset)
            elif offset == 0:
                # log file rotated, nothing to do yet...
                # since there is no first_line
                debug("%s is empty.  File was rotated", logfile)
            elif offset < last_offset:
                # file was rotated or replaced and now has data
                debug("%s most likely rotated and now has data", logfile)
                last_offset = 0
                self.file_tracker.update_first_line()
                continue

            self.sleepAndPurge(daemon_sleep, purge_time,
                               purge_sleep_ratio, sync_sleep_ratio)



    def sleepAndPurge(self, sleep_time, purge_time,
                      purge_sleep_ratio = None, sync_sleep_ratio = None):
        time.sleep(sleep_time)
        if purge_time:
            self.purge_counter += 1
            if self.purge_counter == purge_sleep_ratio:
                try:
                    purge = Purge(self.__prefs,
                                  purge_time)
                except Exception as e:
                    logging.getLogger().exception(e)
                    raise
                self.purge_counter = 0

        if sync_sleep_ratio:
            #debug("sync count: %d", self.sync_counter)
            self.sync_counter += 1
            if self.sync_counter == sync_sleep_ratio:
                try:
                    sync = Sync(self.__prefs)
                    if self.__sync_upload:
                        debug("sync upload")
                        timestamp = sync.send_new_hosts()
                    if self.__sync_download:
                        debug("sync download")
                        new_hosts = sync.receive_new_hosts()
                        if new_hosts:
                            info("received new hosts: %s", str(new_hosts))
                            self.get_denied_hosts()
                            self.update_hosts_deny(new_hosts)
                    sync.xmlrpc_disconnect()
                except Exception as e:
                    logging.getLogger().exception(e)
                    raise
                self.sync_counter = 0


    def get_denied_hosts(self):
        self.__denied_hosts = {}
        for line in open(self.__prefs.get('HOSTS_DENY'), "r"):
            if line[0] not in ('#', '\n'):

                idx = line.find('#')
                if idx != 1:
                    line = line[:idx]
                try:
                    host = parse_host(line)
                    self.__denied_hosts[host] = 0
                    if host in self.__allowed_hosts:
                        self.__allowed_hosts.add_warned_host(host)
                except Exception:
                    pass

        new_warned_hosts = self.__allowed_hosts.get_new_warned_hosts()
        if new_warned_hosts:
            self.__allowed_hosts.save_warned_hosts()

            text = """WARNING: The following hosts appear in %s but should be
allowed based on your %s file"""  % (self.__prefs.get("HOSTS_DENY"),
                                     os.path.join(self.__prefs.get("WORK_DIR"),
                                                  ALLOWED_HOSTS))
            self.__report.add_section(text, new_warned_hosts)
            self.__allowed_hosts.clear_warned_hosts()


    def update_hosts_deny(self, deny_hosts):
        if not deny_hosts: return None, None

        #info("keys: %s", str( self.__denied_hosts.keys()))
        new_hosts = [host for host in deny_hosts
                     if host not in self.__denied_hosts
                     and host not in self.__allowed_hosts]

        debug("new hosts: %s", str(new_hosts))

        try:
            fp = open(self.__prefs.get('HOSTS_DENY'), "a")
            status = 1
        except Exception as e:
            print(e)
            print("These hosts should be manually added to: %s", self.__prefs.get('HOSTS_DENY'))
            # print(self.__prefs.get('HOSTS_DENY'))
            fp = sys.stdout
            status = 0

        write_timestamp = self.__prefs.get('PURGE_DENY') is not None
        for host in new_hosts:
            block_service = self.__prefs.get('BLOCK_SERVICE')
            if block_service:
                block_service = "%s: " % block_service
                output = "%s%s%s" % (block_service, host, BSD_STYLE)
            else:
                output = "%s" % host

            if write_timestamp:
                fp.write("%s %s%s%s\n" % (DENY_DELIMITER,
                                          time.asctime(),
                                          ENTRY_DELIMITER,
                                          output))
            fp.write("%s\n" % output)

        plugin_deny = self.__prefs.get('PLUGIN_DENY')
        if plugin_deny: plugin.execute(plugin_deny, new_hosts)
        if self.__iptables:
           debug("Trying to create iptables rules")
           try:
              for host in new_hosts:
                  my_host = str(host)
                  if self.__blockport:
                     new_rule = self.__iptables + " -I INPUT -p tcp --dport " + self.__blockport + " -s " + my_host + " -j DROP"
                  else:
                     new_rule = self.__iptables + " -I INPUT -s " + my_host + " -j DROP"
                  debug("Running iptabes rule: %s", new_rule)
                  info("Creating new firewall rule %s", new_rule)
                  os.system(new_rule);

           except Exception as e:
               print(e)
               print("Unable to write new firewall rule.")

        if self.__pfctl and self.__pftable:
             debug("Trying to update PF table.")
             try:
               for host in new_hosts:
                   my_host = str(host)
                   new_rule = self.__pfctl + " -t " + self.__pftable + " -T add " + my_host
                   debug("Running PF update rule: %s", new_rule)
                   info("Creating new PF rule %s", new_rule)
                   os.system(new_rule);
                   
             except Exception as e:
                print(e)
                print("Unable to write new PF rule.")
                debug("Unable to create PF rule. %s", e)
        if self.__pftablefile:
              debug("Trying to write host to PF table file %s", self.__pftablefile)
              try:
                 pf_file = open(self.__pftablefile, "a")
                 for host in new_hosts:
                    my_host = str(host)
                    pf_file.write("%s\n" % my_host)
                    info("Wrote new host %s to table file %s", my_host, self.__pftablefile)
                 pf_file.close()
              except Exception as e:
                  print(e)
                  print("Unable to write new host to PF table file.")
                  debug("Unable to write new host to PF table file %s", self.__pftablefile)

        if fp != sys.stdout:
            fp.close()

        return new_hosts, status


    def is_valid(self, rx_match):
        invalid = 0
        try:
            if rx_match.group("invalid"):
                invalid = 1
        except Exception:
            invalid = 1
        return invalid

    def process_log(self, logfile, offset):
        try:
            if logfile.endswith(".gz"):
                fp = gzip.open(logfile)
            elif logfile.endswith(".bz2"):
                if HAS_BZ2: fp = bz2.BZ2File(logfile, "r")
                else:       raise Exception("Can not open bzip2 file (missing bz2 module)")
            else:
                fp = open(logfile, "r")
        except Exception as e:
            print("Could not open log file: %s" % logfile)
            print(e)
            return -1

        try:
            fp.seek(offset)
        except IOError:
            pass

        suspicious_always = is_true(self.__prefs.get('SUSPICIOUS_LOGIN_REPORT_ALLOWED_HOSTS'))

        login_attempt = LoginAttempt(self.__prefs,
                                     self.__allowed_hosts,
                                     suspicious_always,
                                     self.__first_time,
                                     1, # fetch all
                                     self.__restricted)

        for line in fp:
            success = invalid = 0
            m = None
            sshd_m = self.__sshd_format_regex.match(line)
            if sshd_m:
                message = sshd_m.group('message')

                # did this line match any of the fixed failed regexes?
                for i in FAILED_ENTRY_REGEX_RANGE:
                    rx = self.__failed_entry_regex_map.get(i)
                    if rx is None:
                        continue
                    m = rx.search(message)
                    if m:
                        invalid = self.is_valid(m)
                        break
                else: # didn't match any of the failed regex'es, was it succesful?
                    m = self.__successful_entry_regex.match(message)
                    if m:
                        success = 1

            # otherwise, did the line match a failed dovelog login attempt?
            if is_true(self.__prefs.get("DETECT_DOVECOT_LOGIN_ATTEMPTS")):
                rx =  self.__failed_dovecot_entry_regex
                m = rx.search(line)
                if m:
                    #debug("matched (host=%s): %s", m.group("host"), rx.pattern)
                    invalid = self.is_valid(m)

            # otherwise, did the line match one of the userdef regexes?
            if not m:
                for rx in self.__prefs.get('USERDEF_FAILED_ENTRY_REGEX'):
                    m = rx.search(line)
                    if m:
                        #info("matched: %s" % rx.pattern)
                        invalid = self.is_valid(m)
                        break

            if not m:
                # line isn't important
                continue

            try:
                user = m.group("user")
            except Exception:
                user = ""
            try:
                host = m.group("host")
            except Exception:
                error("regex pattern ( %s ) is missing 'host' group" % m.re.pattern)
                continue

            if not is_valid_ip_address(host):
                warning("got invalid host (%s), ignoring" % host)
                continue

            debug ("user: %s - host: %s - success: %d - invalid: %d",
                   user,
                   host,
                   success,
                   invalid)
            login_attempt.add(user, host, success, invalid)

        offset = fp.tell()
        fp.close()

        login_attempt.save_all_stats()
        deny_hosts = login_attempt.get_deny_hosts()

        #print deny_hosts
        new_denied_hosts, status = self.update_hosts_deny(deny_hosts)
        if new_denied_hosts:
            if not status:
                msg = "WARNING: Could not add the following hosts to %s" % self.__prefs.get('HOSTS_DENY')
            else:
                msg = "Added the following hosts to %s" % self.__prefs.get('HOSTS_DENY')
            self.__report.add_section(msg, new_denied_hosts)
            if self.__sync_server: self.sync_add_hosts(new_denied_hosts)
            plugin_deny = self.__prefs.get('PLUGIN_DENY')

            if plugin_deny: plugin.execute(plugin_deny, new_denied_hosts)

        new_suspicious_logins = login_attempt.get_new_suspicious_logins()
        if new_suspicious_logins:
            msg = "Observed the following suspicious login activity"
            self.__report.add_section(msg, list(new_suspicious_logins.keys()))

        if new_denied_hosts:
            info("new denied hosts: %s", str(new_denied_hosts))
        else:
            debug("no new denied hosts")

        if new_suspicious_logins:
            info("new suspicious logins: %s", str(list(new_suspicious_logins.keys())))
        else:
            debug("no new suspicious logins")

        if not self.__report.empty():
            if not self.__noemail:
                # send the report via email if configured
                send_email(self.__prefs, self.__report.get_report())
            elif not self.__daemon:
                # otherwise, if not in daemon mode, log the report to the console
                info(self.__report.get_report())
            self.__report.clear()

        return offset


    def sync_add_hosts(self, hosts):
        try:
            filename = os.path.join(self.__prefs.get("WORK_DIR"), SYNC_HOSTS)
            fp = open(filename, "a")
            for host in hosts:
                fp.write("%s\n" % host)
            fp.close()
            os.chmod(filename, 0o644)
        except Exception as e:
            error(str(e))

    def get_regex(self, name, default):
        val = self.__prefs.get(name)
        if not val:
            return default
        else:
            return re.compile(val)


    def init_regex(self):
        self.__sshd_format_regex = self.get_regex('SSHD_FORMAT_REGEX', SSHD_FORMAT_REGEX)

        self.__successful_entry_regex = self.get_regex('SUCCESSFUL_ENTRY_REGEX',
                                                       SUCCESSFUL_ENTRY_REGEX)

        self.__failed_entry_regex_map = {}
        for i in FAILED_ENTRY_REGEX_RANGE:
            if i == 1: extra = ""
            else: extra = "%i" % i
            self.__failed_entry_regex_map[i] = self.get_regex('FAILED_ENTRY_REGEX%s' % extra,
                                                              FAILED_ENTRY_REGEX_MAP[i])

        self.__failed_dovecot_entry_regex = self.get_regex('FAILED_DOVECOT_ENTRY_REGEX', FAILED_DOVECOT_ENTRY_REGEX)

##        self.__failed_entry_regex = self.get_regex('FAILED_ENTRY_REGEX', FAILED_ENTRY_REGEX)
##        self.__failed_entry_regex2 = self.get_regex('FAILED_ENTRY_REGEX2', FAILED_ENTRY_REGEX2)
##        self.__failed_entry_regex3 = self.get_regex('FAILED_ENTRY_REGEX3', FAILED_ENTRY_REGEX3)
##        self.__failed_entry_regex4 = self.get_regex('FAILED_ENTRY_REGEX4', FAILED_ENTRY_REGEX4)
##        self.__failed_entry_regex5 = self.get_regex('FAILED_ENTRY_REGEX5', FAILED_ENTRY_REGEX5)
##        self.__failed_entry_regex6 = self.get_regex('FAILED_ENTRY_REGEX6', FAILED_ENTRY_REGEX6)
##        self.__failed_entry_regex6 = self.get_regex('FAILED_ENTRY_REGEX7', FAILED_ENTRY_REGEX7)
##        self.__failed_entry_regex6 = self.get_regex('FAILED_ENTRY_REGEX8', FAILED_ENTRY_REGEX8)
##        self.__failed_entry_regex6 = self.get_regex('FAILED_ENTRY_REGEX9', FAILED_ENTRY_REGEX9)
##        self.__failed_entry_regex6 = self.get_regex('FAILED_ENTRY_REGEX10', FAILED_ENTRY_REGEX10)

# vim: set sw=4 et :



"""Disk And Execution MONitor (Daemon)

Default daemon behaviors (they can be modified):
   1.) Ignore SIGHUP signals.
   2.) Default current working directory to the "/" directory.
   3.) Set the current file creation mode mask to 0.
   4.) Close all open files (0 to [SC_OPEN_MAX or 256]).
   5.) Redirect standard I/O streams to "/dev/null".

Failed fork() calls will return a tuple: (errno, strerror).  This behavior
can be modified to meet your program's needs.

Resources:
   Advanced Programming in the Unix Environment: W. Richard Stevens
   Unix Network Programming (Volume 1): W. Richard Stevens
   http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
"""

__author__ = "Chad J. Schroeder"

import os
import signal
import sys

def createDaemon():
    """Detach a process from the controlling terminal and run it in the
    background as a daemon.
    """

    try:
        # Fork a child process so the parent can exit.  This will return control
        # to the command line or shell.  This is required so that the new process
        # is guaranteed not to be a process group leader.  We have this guarantee
        # because the process GID of the parent is inherited by the child, but
        # the child gets a new PID, making it impossible for its PID to equal its
        # PGID.
        pid = os.fork()
    except OSError as e:
        return e.errno, e.strerror  # ERROR (return a tuple)

    if pid == 0:  # The first child.

        # Next we call os.setsid() to become the session leader of this new
        # session.  The process also becomes the process group leader of the
        # new process group.  Since a controlling terminal is associated with a
        # session, and this new session has not yet acquired a controlling
        # terminal our process now has no controlling terminal.  This shouldn't
        # fail, since we're guaranteed that the child is not a process group
        # leader.
        os.setsid()

        # When the first child terminates, all processes in the second child
        # are sent a SIGHUP, so it's ignored.
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        try:
            # Fork a second child to prevent zombies.  Since the first child is
            # a session leader without a controlling terminal, it's possible for
            # it to acquire one by opening a terminal in the future.  This second
            # fork guarantees that the child is no longer a session leader, thus
            # preventing the daemon from ever acquiring a controlling terminal.
            pid = os.fork()  # Fork a second child.
        except OSError as e:
            return e.errno, e.strerror  # ERROR (return a tuple)

        if pid == 0:  # The second child.
            # Ensure that the daemon doesn't keep any directory in use.  Failure
            # to do this could make a filesystem unmountable.
            os.chdir("/")
            # Give the child complete control over permissions.
            os.umask(0)
        else:
            os._exit(0)  # Exit parent (the first child) of the second child.
    else:
        os._exit(0)  # Exit parent of the first child.

    std_fds = 3  # 0,1,2
    for fd in range(0, std_fds):
        try:
            os.close(fd)
        except OSError:  # ERROR (ignore)
            pass

    # Redirect the standard file descriptors to /dev/null.
    os.open("/dev/null", os.O_RDONLY)  # standard input (0)
    os.open("/dev/null", os.O_RDWR)  # standard output (1)
    os.open("/dev/null", os.O_RDWR)  # standard error (2)

    return 0

if __name__ == "__main__":
    # Self-test.

    retCode = createDaemon()

    # If executed with superuser privilages, there should be a new file in the
    # "/" directory.  It should contain the function's return code, the daemon's
    # PID, PPID, and PGRP.  Its PID should not equal its PGRP, and its PPID
    # should equal 1.  If it's executed without superuser privilages, the file
    # won't be created and no errors will be reported.
    open("createDaemon.log", "w").write("rc: %s; pid: %d; ppid: %d; pgrp: %d" % \
                                        (retCode, os.getpid(), os.getppid(), os.getpgrp()))

    sys.exit(0)



from __future__ import print_function, unicode_literals

import unittest

from DenyHosts.util import is_false, is_true, parse_host

# List of 2-tuples: line to parse, expected host
HOST_TEST_DATA = [
    ('host', 'host'),
    ('ALL:127.0.0.1', '127.0.0.1'),
    (3, ''),
]

class UtilsTest(unittest.TestCase):
    def setUp(self):
        self.true_strings = ['1', 't', 'true', 'y', 'yes']
        self.false_strings = ['', 'false', 'ye', 'tr']

    def test_is_true(self):
        for string in self.true_strings:
            self.assertTrue(is_true(string))

        for string in self.false_strings:
            self.assertFalse(is_true(string))

    def test_is_false(self):
        for string in self.true_strings:
            self.assertFalse(is_false(string))

        for string in self.false_strings:
            self.assertTrue(is_false(string))

    def test_parse_host(self):
        for line, expected in HOST_TEST_DATA:
            self.assertEqual(parse_host(line), expected)



import os

from .constants import RESTRICTED_USERNAMES

class Restricted:
    def __init__(self, prefs):
        self.filename = os.path.join(prefs['ETC_DIR'], RESTRICTED_USERNAMES)
        self.__data = set()
        self.load_restricted()

    def load_restricted(self):
        try:
            fp = open(self.filename, "r")
            for line in fp:
                line = line.strip()
                if not line: continue
                if line[0] == '#': continue
                self.__data.add(line)
        except IOError:
            pass

    def get_restricted(self):
        return self.__data



import logging
import os

from . import constants
from .counter import Counter, CounterRecord

error = logging.getLogger("purgecounter").error
info = logging.getLogger("purgecounter").info

class PurgeCounter(object):
    def __init__(self, prefs):
        self.filename = os.path.join(prefs['WORK_DIR'],
                                     constants.PURGE_HISTORY)
        self.purge_threshold = prefs['PURGE_THRESHOLD']


    def get_banned_for_life(self):
        banned = set()
        if self.purge_threshold == 0:
            return banned

        try:
            fp = open(self.filename, "r")
        except IOError:
            return banned

        for line in fp:
            try:
                host, count, timestamp = line.strip().split(':', 2)
            except Exception:
                continue

            if int(count) > self.purge_threshold:
                banned.add(host)

        fp.close()
        return banned


    def get_data(self):
        counter = Counter()
        try:
            fp = open(self.filename, "r")
        except IOError:
            return counter

        for line in fp:
            try:
                host, count, timestamp = line.strip().split(':', 2)
            except Exception:
                continue
            counter[host] = CounterRecord(int(count), timestamp)

        fp.close()
        return counter


    def write_data(self, data):
        try:
            fp = open(self.filename, "w")
            keys = list(data.keys())
            keys.sort()

            for key in keys:
                fp.write("%s:%s\n" % (key, data[key]))
            fp.close()
        except Exception as e:
            error("error saving %s: %s", self.filename, str(e))

    def increment(self, purged_hosts):
        data = self.get_data()

        for host in purged_hosts:
            data[host] += 1
        self.write_data(data)



import logging
import os

error = logging.getLogger("plugin").error
info = logging.getLogger("plugin").info
debug = logging.getLogger("plugin").debug

def execute(executable, hosts):
    for host in hosts:
        debug("invoking plugin: %s %s", executable, host)
        try:
            res = os.system("%s %s" % (executable, host))
            if res: info("plugin returned %d", res)
        except Exception as e:
            error("plugin error: %s", e)



#!/usr/bin/env python
# Copyright 2005-2006 (C) Phil Schwartz <phil_schwartz@users.sourceforge.net>
# Copyright 2014 (C) Jesse Smith <jessefrgsmith@yahoo.ca>

from glob import glob
from os.path import join as ospj

from distutils.core import setup

from DenyHosts.util import normalize_whitespace
from DenyHosts.version import VERSION

etcpath = "/etc"
manpath = "/usr/share/man/man8"
libpath = "/usr/share/denyhosts"
scriptspath = ospj("scripts", libpath)
pluginspath = ospj("plugins", libpath)

setup(
    name="DenyHosts",
    version=VERSION,
    description="DenyHost is a utility to help sys admins thwart ssh hackers",
    author="Jesse Smith",
    author_email="jessefrgsmith@yahoo.ca",
    url="http://denyhost.sourceforge.net",
    scripts=['denyhosts.py', 'daemon-control-dist'],
    package_dir={'DenyHosts': 'DenyHosts'},
    packages=["DenyHosts"],
    requires=["ipaddr"],
    data_files=[
        (etcpath, glob("denyhosts.conf")),
        (manpath, glob("denyhosts.8")),
    ],
    license="GPL v2",
    long_description=normalize_whitespace(
        """
        DenyHosts is a python program that automatically blocks ssh attacks
        by adding entries to /etc/hosts.deny. DenyHosts will also inform
        administrators about offending hosts, attacked users and suspicious
        logins. Originally written by Phil Schwartz.
        """
    ),
)



import os
import sys
from .util import die

class LockFile(object):
    def __init__(self, lockpath):
        self.lockpath = lockpath
        self.fd = None

    def exists(self):
        return os.access(self.lockpath, os.F_OK)

    def get_pid(self):
        pid = ""
        try:
            fp = open(self.lockpath, "r")
            pid = fp.read().strip()
            fp.close()
        except IOError:
            pass
        return pid

    def create(self):
        try:
            self.fd = os.open(self.lockpath,
                              os.O_CREAT |  # create file
                              os.O_TRUNC |  # truncate it, if it exists
                              os.O_WRONLY | # write-only
                              os.O_EXCL,    # exclusive access
                              0o644)         # file mode

        except Exception as e:
            pid = self.get_pid()
            die("DenyHosts could not obtain lock (pid: %s)" % pid, e)

        s = "%s\n" % os.getpid()
        if sys.version_info < (3, 0):
             os.write(self.fd, s)
        else:
             os.write(self.fd, bytes(s, 'UTF-8'))
        os.fsync(self.fd)

    def remove(self, die_=True):
        try:
            if self.fd:
                os.close(self.fd)
        except IOError:
            pass

        self.fd = None
        try:
            os.unlink(self.lockpath)
        except Exception as e:
            if die_:
                die("Error deleting DenyHosts lock file: %s" % self.lockpath, e)



#!/usr/bin/env python
#
############################################################################
# this script will read the /etc/passwd file and extract usernames
# that have one of the RESTRICTED_SHELLS.  The ouput of this
# script is a list of these restricted usernames suitable for
# use in WORK_DIR/restricted-usernames
#
# such as: python restricted_from_passwd > $WORK_DIR/restricted-usernames
# where $WORK_DIR is your DenyHosts WORK_DIR parameter
#
############################################################################

RESTRICTED_SHELLS = ("/sbin/nologin",
                     "/usr/sbin/nologin",
                     "/sbin/shutdown",
                     "/sbin/halt")

from pwd import getpwall

passwd = getpwall()

usernames = []
for pw_tuple in passwd:
    if pw_tuple[6] in RESTRICTED_SHELLS:
        usernames.append(pw_tuple[0])

usernames.sort()
for username in usernames:
    print username





from __future__ import print_function, unicode_literals

from os.path import dirname, join as ospj
from tempfile import mkdtemp
import unittest

from DenyHosts.prefs import Prefs
from DenyHosts.restricted import Restricted

class EmptyRestrictedTest(unittest.TestCase):
    """
    Tests creating a Restricted object with no data file. This should:
    1) not throw an exception, and
    2) result in an empty set of restricted hosts.

    We ensure that there's no data file by making an empty temporary
    directory for this test, and assinging that path to the appropriate
    Prefs key.
    """
    def setUp(self):
        self.prefs = Prefs()
        self.prefs._Prefs__data['ETC_DIR'] = mkdtemp()

    def test_no_data_file(self):
        restricted = Restricted(self.prefs)
        self.assertFalse(restricted.get_restricted())

class RestrictedTest(unittest.TestCase):
    """
    Tests creating a Restricted object with a data file.
    """
    def setUp(self):
        self.prefs = Prefs()
        self.prefs._Prefs__data['ETC_DIR'] = ospj(dirname(__file__), 'data/restricted')
        self.test_hosts = set(['host1', 'host2'])

    def test_no_data_file(self):
        restricted = Restricted(self.prefs)
        self.assertEqual(restricted.get_restricted(), self.test_hosts)
